# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing utilities
=================
"""
from typing import Any
import os
import pathlib
import platform

import fsspec
import fsspec.implementations.local

from .. import fs_utils
# pylint: disable=unused-import # Need to import for fixtures
from .cluster import dask_client, dask_cluster

# pylint: disable=unused-import

#: Test data
TEXT = '''Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam porta
turpis dictum, porta tellus eu, convallis mi. Integer at placerat diam. Donec in
various neque. Morbi sed nisi finibus, mattis velit non, pulvinar metus. Duis
feugiat diam eget augue posuere, nec aliquam dolor tristique. Aliquam a dolor
vel ante sagittis dictum vel at dolor. Suspendisse velit dolor, vestibulum eget
aliquet ut, imperdiet at justo. Nullam sit amet suscipit orci, bibendum sagittis
orci. Aliquam mattis feugiat rutrum. Vivamus fermentum ex non mauris faucibus
vehicula. Donec odio lacus, viverra et hendrerit eu, mollis eget mauris. Duis
suscipit, velit nec finibus ullamcorper, nisi lorem fermentum tellus, ut viverra
nunc lorem ut odio. Duis eget ligula maximus, venenatis nulla a, commodo dolor.
Aenean justo sapien, mollis aliquam vestibulum id, suscipit a ligula. Phasellus
porta arcu erat, elementum faucibus leo auctor vel. Integer vel pharetra leo.'''


def test_join_path() -> None:
    """Test the join_path function."""
    assert fs_utils.join_path('a', 'b', 'c') == 'a/b/c'
    assert fs_utils.join_path('a', 'b', 'c', 'd') == 'a/b/c/d'
    assert fs_utils.join_path('a', 'b', 'c', 'd', 'e') == 'a/b/c/d/e'
    assert fs_utils.join_path('a', 'b', 'c', 'd', 'e', 'f') == 'a/b/c/d/e/f'


def test_get_fs() -> None:
    """Test the get_fs function."""
    fs = fs_utils.get_fs('file')
    assert isinstance(fs, fsspec.implementations.local.LocalFileSystem)
    fs = fs_utils.get_fs()
    assert isinstance(fs, fsspec.implementations.local.LocalFileSystem)


def test_fs_walk(tmpdir) -> None:
    """Test the fs_walk function."""
    item: Any

    for idx, item in enumerate([
        ('year=2014', 'month=5'),
        ('year=2014', 'month=5', 'day=2'),
        ('year=2014', 'month=5', 'day=1'),
        ('year=2014', 'month=5', 'day=3'),
        ('year=2014', 'month=4'),
        ('year=2014', 'month=4', 'day=16'),
        ('year=2014', 'month=4', 'day=24'),
        ('year=2014', 'month=4', 'day=27'),
        ('year=2014', 'month=4', 'day=20'),
        ('year=2014', 'month=4', 'day=29'),
        ('year=2014', 'month=4', 'day=14'),
        ('year=2014', 'month=4', 'day=25'),
        ('year=2014', 'month=4', 'day=19'),
        ('year=2014', 'month=4', 'day=12'),
        ('year=2014', 'month=4', 'day=23'),
        ('year=2014', 'month=4', 'day=17'),
        ('year=2014', 'month=4', 'day=28'),
        ('year=2014', 'month=4', 'day=13'),
        ('year=2014', 'month=4', 'day=21'),
        ('year=2014', 'month=4', 'day=15'),
        ('year=2014', 'month=4', 'day=18'),
        ('year=2014', 'month=4', 'day=26'),
        ('year=2014', 'month=4', 'day=22'),
        ('year=2014', 'month=4', 'day=30'),
    ]):
        path = pathlib.Path(tmpdir).joinpath(*item)
        path.mkdir(parents=True, exist_ok=False)
        if 'day' in item[-1]:
            with path.joinpath(f'file_{idx}.txt').open(mode='w',
                                                       encoding='utf-8'):
                ...

    fs = fs_utils.get_fs()
    listing1 = []
    for root, _dirs, files in fs_utils.fs_walk(fs, str(tmpdir), sort=True):
        for item in files:
            listing1.append(fs.sep.join([root, item]))

    listing2 = []
    for root, _dirs, files in fs_utils.fs_walk(fs, str(tmpdir), sort=False):
        for item in files:
            listing2.append(fs.sep.join([root, item]))

    assert listing1 == sorted(listing2)

    assert list(
        fs_utils.fs_walk(fs,
                         str(pathlib.Path(tmpdir).joinpath('inexistent')),
                         sort=True)) == [('', [], [])]


def test_normalize_path() -> None:
    """Test the normalize_path function."""
    fs = fsspec.filesystem('file')
    root = str(pathlib.Path('/').resolve())
    if platform.system() == 'Windows':
        # fsspec returns only the drive letter for the root path.
        root = root.replace('\\', '')

    def istrcmp(str1, str2):
        """Case insensitive string comparison."""
        return str1.lower() == str2.lower()

    assert istrcmp(fs_utils.normalize_path(fs, '/'), root)
    assert istrcmp(fs_utils.normalize_path(fs, './foo'),
                   str(pathlib.Path('.').resolve() / 'foo'))

    fs = fsspec.filesystem('memory')
    assert fs_utils.normalize_path(fs, '/') == os.path.sep
    assert fs_utils.normalize_path(fs, './foo') == f'{os.path.sep}foo'

    fs = fsspec.filesystem('s3')
    assert fs_utils.normalize_path(fs, '/') == '/'
    assert fs_utils.normalize_path(fs, './foo') == './foo'


def test_copy_file(tmpdir) -> None:
    """Test the copy file across different file systems."""
    fs_source = fsspec.filesystem('file')
    fs_target = fsspec.filesystem('memory')
    path = str(tmpdir / 'foo.txt')
    with fs_source.open(path, mode='wb', encoding='utf-8') as stream:
        stream.write(TEXT.encode('utf-8'))
    fs_utils.copy_file(path, 'foo.txt', fs_source, fs_target)

    assert fs_target.cat('foo.txt').decode('utf-8') == TEXT


def test_copy_files(tmpdir) -> None:
    """Test the copy files across different file systems."""
    source = tmpdir / 'source'
    target = tmpdir / 'target'
    fs_source = fsspec.filesystem('file')
    fs_target = fsspec.filesystem('file')
    fs_source.mkdir(source)
    fs_target.mkdir(target)
    paths = [
        str(source / item) for item in (
            'foo.txt',
            'bar.txt',
            'baz.txt',
        )
    ]
    for path in paths:
        with fs_source.open(path, mode='wb', encoding='utf-8') as stream:
            stream.write(TEXT.encode('utf-8'))
    fs_utils.copy_files(paths, str(target), fs_source, fs_target)

    for item in fs_target.ls(str(target)):
        assert fs_target.cat(item).decode('utf-8') == TEXT


def test_copy_tree(tmpdir) -> None:
    """Test the copy tree across different file systems."""
    item: Any
    fs_source = fsspec.filesystem('file')
    fs_target = fsspec.filesystem('memory')

    for idx, item in enumerate([
        ('year=2014', 'month=5'),
        ('year=2014', 'month=5', 'day=2'),
        ('year=2014', 'month=5', 'day=1'),
        ('year=2014', 'month=5', 'day=3'),
        ('year=2014', 'month=4'),
        ('year=2014', 'month=4', 'day=16'),
        ('year=2014', 'month=4', 'day=24'),
        ('year=2014', 'month=4', 'day=27'),
        ('year=2014', 'month=4', 'day=20'),
        ('year=2014', 'month=4', 'day=29'),
        ('year=2014', 'month=4', 'day=14'),
        ('year=2014', 'month=4', 'day=25'),
        ('year=2014', 'month=4', 'day=19'),
        ('year=2014', 'month=4', 'day=12'),
        ('year=2014', 'month=4', 'day=23'),
        ('year=2014', 'month=4', 'day=17'),
        ('year=2014', 'month=4', 'day=28'),
        ('year=2014', 'month=4', 'day=13'),
        ('year=2014', 'month=4', 'day=21'),
        ('year=2014', 'month=4', 'day=15'),
        ('year=2014', 'month=4', 'day=18'),
        ('year=2014', 'month=4', 'day=26'),
        ('year=2014', 'month=4', 'day=22'),
        ('year=2014', 'month=4', 'day=30'),
    ]):
        path = fs_utils.join_path(str(tmpdir), *item)
        fs_source.makedirs(path, exist_ok=False)
        if 'day' in item[-1]:
            with fs_source.open(fs_utils.join_path(path, f'file_{idx}.txt'),
                                mode='wb',
                                encoding='utf-8') as stream:
                stream.write(TEXT.encode('utf-8'))

    fs_utils.copy_tree(str(tmpdir), '/tree', fs_source, fs_target)

    for root, dirs, files in fs_utils.fs_walk(fs_target, '/tree'):
        for item in files:
            assert fs_target.cat(fs_utils.join_path(
                root, item)).decode('utf-8') == TEXT
        for item in dirs:
            item = item.replace('\\', '/')
            parts = item.replace('/tree/', '').split(fs_target.sep)
            assert parts[0] == 'year=2014'
            if len(parts) > 1:
                assert parts[1] in ['month=4', 'month=5']
            if len(parts) > 2:
                assert 'day=' in parts[2]
            if len(parts) > 3:
                assert 'file_' in parts[3]
                assert parts[3].endswith('.txt')



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Test merging.
=============
"""
import numpy
import pytest
import zarr

from .. import _update_fs, merge_time_series, perform
from ... import sync
from ...tests import data
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster
from ...tests.fixture import dask_arrays, numpy_arrays
from ...tests.fs import local_fs

# pylint: enable=unused-import


class MyError(RuntimeError):
    """Custom error."""


class ThrowError(sync.Sync):
    """Throw an error when merging."""

    def __enter__(self) -> bool:
        raise MyError('This is an error')

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        ...

    def is_locked(self) -> bool:
        return False


def test_update_fs(
        dask_client,  # pylint: disable=redefined-outer-name
        local_fs,  # pylint: disable=redefined-outer-name
) -> None:
    """Test the _update_fs function."""
    generator = data.create_test_dataset(delayed=False)
    zds = next(generator)

    partition_folder = local_fs.root.joinpath('partition_folder')

    zattrs = str(partition_folder.joinpath('.zattrs'))
    future = dask_client.submit(_update_fs, str(partition_folder),
                                dask_client.scatter(zds), local_fs.fs)
    dask_client.gather(future)
    assert local_fs.exists(zattrs)

    local_fs.fs.rm(str(partition_folder), recursive=True)
    assert not local_fs.exists(zattrs)
    seen_exception = False
    try:
        future = dask_client.submit(_update_fs,
                                    str(partition_folder),
                                    dask_client.scatter(zds),
                                    local_fs.fs,
                                    synchronizer=ThrowError())
        dask_client.gather(future)
    except MyError:
        seen_exception = True
    assert seen_exception
    assert not local_fs.exists(zattrs)


@pytest.mark.parametrize('arrays_type', ['dask_arrays', 'numpy_arrays'])
def test_perform(
    dask_client,  # pylint: disable=redefined-outer-name
    local_fs,  # pylint: disable=redefined-outer-name
    arrays_type,
    request,
) -> None:
    """Test the perform function."""
    delayed = request.getfixturevalue(arrays_type)
    generator = data.create_test_dataset(delayed=delayed)
    zds = next(generator)

    path = str(local_fs.root.joinpath('folder'))

    future = dask_client.submit(_update_fs, path, dask_client.scatter(zds),
                                local_fs.fs)
    dask_client.gather(future)

    future = dask_client.submit(perform,
                                dask_client.scatter(zds),
                                path,
                                'time',
                                local_fs.fs,
                                'time',
                                delayed=delayed,
                                merge_callable=merge_time_series)
    dask_client.gather(future)

    zgroup = zarr.open_consolidated(local_fs.get_mapper(path))
    assert numpy.all(zgroup['time'][...] == zds['time'].values)
    assert numpy.all(zgroup['var1'][...] == zds['var1'].values)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Time period
===========
"""
from __future__ import annotations

from typing import Any, Callable, Match
import enum
import re

import numpy

from ..type_hints import DType

# Parse the unit of numpy.timedelta64.
PATTERN: Callable[[str], Match[str] | None] = re.compile(
    r'(?:datetime|timedelta)64\[(\w+)\]').search

#: Numpy time units
RESOLUTION = ('as', 'fs', 'ps', 'ns', 'us', 'ms', 's', 'm', 'h', 'D', 'W', 'M',
              'Y')


def _time64_unit(dtype: DType[Any]) -> str:
    """Get the unit of time."""
    match: Match[str] | None = PATTERN(dtype.name)
    if match is None:
        raise ValueError(f'dtype is not a time duration: {dtype}')
    return match.group(1)


def _min_time64_unit(*args: DType[Any]) -> str:
    """Get the minimum unit of time."""
    index = min(RESOLUTION.index(_time64_unit(item)) for item in args)
    return RESOLUTION[index]


class PeriodRelation(enum.IntEnum):
    """Enumeration of the relations which can exist between two periods."""
    __slots__ = ()

    #: The first period is after.
    #:
    #: .. code-block:: text
    #:
    #:             [-----p1-----]
    #:      [-p2-]
    AFTER = 0

    #: The first period is after, but the start date of the first period is
    #: the same date as the end date of the second period.
    #:
    #: .. code-block:: text
    #:
    #:           [-----p1-----]
    #:      [-p2-]
    START_TOUCHING = 1

    #: The start date of the first period is inside the second period.
    #:
    #: .. code-block:: text
    #:
    #:        [-----p1-----]
    #:      [-p2-]
    START_INSIDE = 2

    #: The start date of the first period is same as the first date of the
    #: second period and the end date is inside the first period.
    #:
    #: .. code-block:: text
    #:
    #:      [-----p1-----]
    #:      [-------p2-------]
    INSIDE_START_TOUCHING = 3

    #: The first period is inside the second period and the start date of the
    #: two periods are the same.
    #:
    #: .. code-block:: text
    #:
    #:      [-----p1-----]
    #:      [--p2--]
    ENCLOSING_START_TOUCHING = 4

    #: The second period is inside the first period.
    #:
    #: .. code-block:: text
    #:
    #:      [-----p1-----]
    #:         [--p2--]
    ENCLOSING = 5

    #: The second period is inside the first period and the end date of the
    #: two periods are the same.
    #:
    #: .. code-block:: text
    #:
    #:      [-----p1-----]
    #:            [--p2--]
    ENCLOSING_END_TOUCHING = 6

    #: The two periods are exactly the same.
    #:
    #: .. code-block:: text
    #:
    #:      [-----p1-----]
    #:      [-----p2-----]
    EXACT_MATCH = 7

    #: The first period is inside the second period.
    #:
    #: .. code-block:: text
    #:
    #:        [---p1---]
    #:      [-----p2-----]
    INSIDE = 8

    #: The first period is inside the second period and the end date of the
    #: two periods are the same.
    #:
    #: .. code-block:: text
    #:
    #:          [---p1---]
    #:      [-----p2-----]
    INSIDE_END_TOUCHING = 9

    #: The end date of the first period is inside the second period.
    #:
    #: .. code-block:: text
    #:
    #:      [---p1---]
    #:            [-----p2-----]
    END_INSIDE = 10

    #: The end date of the first period is same as the start date of the
    #: second period.
    #:
    #: .. code-block:: text
    #:
    #:      [---p1---]
    #:               [-----p2-----]
    END_TOUCHING = 11

    #: The first period is before the second period.
    #:
    #: .. code-block:: text
    #:
    #:      [---p1---]
    #:                  [-----p2-----]
    BEFORE = 12

    def is_after(self) -> bool:
        """Return true if the relation is after."""
        # pylint: disable=comparison-with-callable
        return self.value == PeriodRelation.AFTER
        # pylint: enable=comparison-with-callable

    def is_before(self) -> bool:
        """Return true if the relation is before."""
        # pylint: disable=comparison-with-callable
        return self.value == PeriodRelation.BEFORE
        # pylint: enable=comparison-with-callable

    def contains(self) -> bool:
        """Return true if one period enclosing the other period."""
        return self.value in (PeriodRelation.ENCLOSING_END_TOUCHING,
                              PeriodRelation.ENCLOSING_START_TOUCHING,
                              PeriodRelation.ENCLOSING,
                              PeriodRelation.EXACT_MATCH)

    def is_before_overlapping(self) -> bool:
        """Return true if one period is before but there is an overlap between
        the two periods."""
        return self.value in (PeriodRelation.END_INSIDE,
                              PeriodRelation.END_TOUCHING,
                              PeriodRelation.INSIDE_START_TOUCHING)

    def is_after_overlapping(self) -> bool:
        """Return true if one period is after but there is an overlap between
        the two periods."""
        return self.value in (PeriodRelation.START_INSIDE,
                              PeriodRelation.START_TOUCHING,
                              PeriodRelation.INSIDE_END_TOUCHING)

    def is_inside(self) -> bool:
        """Return true if one period is inside the other period."""
        # pylint: disable=comparison-with-callable
        return self.value == PeriodRelation.INSIDE
        # pylint: enable=comparison-with-callable


class Period:
    """Create a Period from begin to last eg: [begin, last[

    Args:
        begin: The beginning of the period.
        end: The ending of the period.
        within: If true, the given period defines a closed interval
            (i.e. the end date is within the period), otherwise the
            interval is open.
    """
    __slots__: tuple[str, ...] = ('_begin', '_duration_unit', '_last')

    def __init__(self,
                 begin: numpy.datetime64,
                 end: numpy.datetime64,
                 *,
                 within: bool = False) -> None:
        duration_unit: str = _min_time64_unit(begin.dtype, end.dtype)

        #: The beginning of the period.
        self._begin: numpy.datetime64 = begin
        #: The duration unit of the period.
        self._duration_unit: numpy.timedelta64 = numpy.timedelta64(
            1, duration_unit)
        #: The last date of the period.
        self._last: numpy.datetime64 = (end if within else end -
                                        self._duration_unit)

    def __repr__(self) -> str:
        return f'[{self._begin}, {self.end()}['

    def __getstate__(self) -> tuple[Any, ...]:
        return self._begin, self._duration_unit, self._last

    def __setstate__(self, state: tuple[Any, ...]) -> None:
        self._begin, self._duration_unit, self._last = state

    @classmethod
    def from_duration(cls, begin: numpy.datetime64,
                      duration: numpy.timedelta64) -> Period:
        """Create a Period as [begin, begin + duration[

        Args:
            begin: The beginning of the period.
            duration: The duration of the period.

        Returns:
            The created period.
        """
        return cls(begin, begin + duration)

    @property
    def begin(self) -> numpy.datetime64:
        """Return the first element in the period."""
        return self._begin

    @property
    def last(self) -> numpy.datetime64:
        """Return the last item in the period."""
        return self._last

    def end(self) -> numpy.datetime64:
        """Return one past the last element."""
        return self._last + self._duration_unit

    def is_null(self) -> bool:
        """True if period is ill formed (length is zero or less)"""
        return bool(self.end() <= self._begin)  # numpy.bool_ -> bool

    def length(self) -> numpy.timedelta64:
        """Return the length of the period."""
        if self._last < self._begin:
            # invalid period
            return self._last + self._duration_unit - self._begin
        return self.end() - self._begin

    def __eq__(self, rhs: Any) -> bool:
        if not isinstance(rhs, Period):
            return NotImplemented
        return (self._begin == rhs._begin) and (self._last == rhs._last)

    def __ne__(self, rhs: Any) -> bool:
        if not isinstance(rhs, Period):
            return NotImplemented
        return (self._begin != rhs._begin) or (self._last != rhs._last)

    def __lt__(self, rhs: Any) -> bool:
        if not isinstance(rhs, Period):
            return NotImplemented
        return bool(self._last < rhs._begin)  # numpy.bool_ -> bool

    def shift(self, duration: numpy.timedelta64) -> None:
        """Shift the start and end by the specified amount.

        Args:
            duration: The amount to shift the period.
        """
        self._begin = self._begin + duration
        self._last = self._last + duration

    def expand(self, duration: numpy.timedelta64) -> None:
        """Expand the size of the period by the duration on both ends.

        So before expand:

        .. code-block:: text

                    [-------]
            ^   ^   ^   ^   ^   ^  ^
            1   2   3   4   5   6  7

        After expand(2):

        .. code-block:: text

            [----------------------]
            ^   ^   ^   ^   ^   ^  ^
            1   2   3   4   5   6  7

        Args:
            duration: The amount to expand the period.
        """
        self._begin = self._begin - duration
        self._last = self._last + duration

    def contains(
        self,
        other: numpy.datetime64 | Period,
    ) -> bool:
        """Check if the given period is contains this period.

        Args:
            other: The other period to check.

        Returns:
            * True if other is a date and is inside the period, zero length
              periods contain no points
            * True if other is a period and  fully contains (or equals) the
              other period
        """
        if isinstance(other, numpy.datetime64):
            return bool(self._begin <= other <= self._last)
        return bool((self._begin <= other.begin)
                    and (self._last >= other.last))

    def is_adjacent(self, other: Period) -> bool:
        """True if periods are next to each other without a gap.

        In the example below, p1 and p2 are adjacent, but p3 is not adjacent
        with either of p1 or p2.

        .. code-block:: text

            [-p1-[
                 [-p2-]
                   [-p3-]

        Args:
            other: The other period to check.

        Returns:
            * True if other is a date and is adjacent to this period
            * True if other is a period and is adjacent to this period
        """
        return other.begin == self.end() or self.begin == other.end()

    def is_after(self, point: numpy.datetime64) -> bool:
        """True if all of the period is prior or point < start.

        In the example below only point 1 would evaluate to true.

        .. code-block:: text

                [----------[
            ^   ^    ^     ^   ^
            1   2    3     4   5

        Args:
            point: The point to check.

        Returns:
            True if point is after the period
        """
        if self.is_null():
            # null period isn't after
            return False
        return bool(point < self._begin)

    def is_before(self, point: numpy.datetime64) -> bool:
        """True if all of the period is prior to the passed point or end <=
        point.

        In the example below points 4 and 5 return true.

        .. code-block:: text

                [----------[
            ^   ^    ^     ^   ^
            1   2    3     4   5

        Args:
            point: The point to check.

        Returns:
            True if point is before the period
        """
        if self.is_null():
            # null period isn't before anything
            return False
        return bool(self._last < point)

    def intersects(self, other: Period) -> bool:
        """True if the periods overlap in any way.

        In the example below p1 intersects with p2, p4, and p6.

        .. code-block:: text

                  [---p1---[
                        [---p2---[
                           [---p3---[
             [---p4---[
            [-p5-[
                    [-p6-[

        Args:
            other: The other period to check.

        Returns:
            True if the periods intersect
        """
        return bool(
            self.contains(other.begin) or other.contains(self._begin)
            or ((other.begin < self._begin) and (other.last >= self._begin)))

    def intersection(self, other: Period) -> Period:
        """Return the period of intersection or null period if no intersection.

        Args:
            other: The other period to check.

        Returns:
            The intersection period or null period if no intersection.
        """
        if self._begin > other.begin:
            if self._last <= other.last:
                return self
            return Period(self._begin, other.end())

        if self._last <= other.last:
            return Period(other.begin, self.end())
        return other

    def merge(self, other: Period) -> Period:
        """Return the union of intersecting periods -- or null period.

        Args:
            other: The other period to merge.

        Returns:
            The union period of intersection or null if no intersection.
        """
        if self.intersects(other):
            if self._begin < other.begin:
                return Period(
                    self._begin,
                    self.end() if self._last > other.last else other.end())
            return Period(
                other.begin,
                self.end() if self._last > other.last else other.end())
        # no intersect return null
        return Period(self._begin, self._begin)

    def span(self, other: Period) -> Period:
        """Combine two periods with earliest start and latest end.

        Combines two periods and any gap between them such that
          * start = min(p1.start, p2.start)
          * end   = max(p1.end  , p2.end)

        .. code-block:: text

            [---p1---]
                           [---p2---]

        result:

        .. code-block:: text

            [-----------p3----------]

        Args:
            other: The other period to combine.

        Returns:
            The combined period.
        """
        start: numpy.datetime64 = (self._begin if self._begin < other.begin
                                   else other.begin)
        end: numpy.datetime64 = (other.end()
                                 if self._last < other.last else self.end())
        return Period(start, end)

    # pylint: disable=too-many-return-statements
    # This code has a lot of "return" statements because we have to determine
    # the relation between the two periods among 8 different cases.
    def _get_direct_relation(self, other: Period) -> PeriodRelation | None:
        """Get the direct relation between two periods."""
        if other.last < self._begin:
            return PeriodRelation.AFTER

        if other.begin > self._last:
            return PeriodRelation.BEFORE

        if other.begin == self._begin and other.last == self._last:
            return PeriodRelation.EXACT_MATCH

        if other.last == self._begin:
            return PeriodRelation.START_TOUCHING

        if other.begin == self._last:
            return PeriodRelation.END_TOUCHING

        if self.contains(other):
            if other.begin == self.begin:
                return PeriodRelation.ENCLOSING_START_TOUCHING
            return (PeriodRelation.ENCLOSING_END_TOUCHING
                    if other.last == self.last else PeriodRelation.ENCLOSING)

        return None
        # pylint: enable=too-many-return-statements

    def get_relation(self, other: Period) -> PeriodRelation:
        """Get the relationship between the two time periods.

        Args:
            other: Period to consider

        Returns:
            The relation.
        """
        relation: PeriodRelation | None = self._get_direct_relation(other)
        if relation is not None:
            return relation

        period_contains_start: bool = other.contains(self._begin)
        period_contains_end: bool = other.contains(self._last)

        if period_contains_start and period_contains_end:
            if other.begin == self.begin:
                return PeriodRelation.INSIDE_START_TOUCHING
            return (PeriodRelation.INSIDE_END_TOUCHING
                    if other.last == self.last else PeriodRelation.INSIDE)

        if period_contains_start:
            return PeriodRelation.START_INSIDE

        assert period_contains_end, 'Period must contain end'
        return PeriodRelation.END_INSIDE



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Merging a time series
=====================
"""
import numpy

from . import period
from .. import dataset
from ..type_hints import NDArray


def merge_time_series(
    existing_ds: dataset.Dataset,
    inserted_ds: dataset.Dataset,
    axis: str,
    partitioning_dim: str,
) -> dataset.Dataset:
    """Merge two time series together.

    Replaces only the intersection between the existing dataset and the new one,
    but also keeps the existing records if they have not been updated.

    The following figure illustrates the implemented algorithm. Column ``A``
    represents the new data and column ``B``, the data already present. The
    different cells in the columns represent the hours on the day of the
    measurements. The merge result is shown in column ``C``. It contains the
    measurements of the column ``A`` or column ``B`` if column ``A`` does not
    replace them.

    .. figure:: ../images/merge_time_series.svg
        :align: center
        :width: 50%

    Args:
        existing_ds: The existing dataset.
        inserted_ds: The inserted dataset.
        axis: The axis to merge on.
        partitioning_dim: The name of the partitioning dimension.

    Returns:
        The merged dataset.
    """
    existing_axis: NDArray = existing_ds.variables[axis].values
    inserted_axis: NDArray = inserted_ds.variables[axis].values
    existing_period = period.Period(existing_axis.min(),
                                    existing_axis.max(),
                                    within=True)
    inserted_period = period.Period(inserted_axis.min(),
                                    inserted_axis.max(),
                                    within=True)

    relation: period.PeriodRelation = inserted_period.get_relation(
        existing_period)

    # The new piece is located before the existing data.
    if relation.is_before():
        return inserted_ds.concat(existing_ds, partitioning_dim)

    # The new piece is located after the existing data.
    if relation.is_after():
        return existing_ds.concat(inserted_ds, partitioning_dim)

    # The new piece replace the old one
    if relation.contains():
        return inserted_ds

    intersection: period.Period = inserted_period.intersection(existing_period)

    # The new piece is located before, but there is an overlap
    # between the two datasets.
    if relation.is_before_overlapping():
        # pylint: disable=comparison-with-callable
        indices = numpy.where(
            # comparison between ndarray and datetime64
            existing_axis > intersection.end())[0]  # type: ignore
        # pylint: enable=comparison-with-callable
        return inserted_ds.concat(
            existing_ds.isel({partitioning_dim: indices}), partitioning_dim)

    # The new piece is located after, but there is an overlap
    # between the two datasets.
    if relation.is_after_overlapping():
        # pylint: disable=comparison-with-callable
        indices = numpy.where(
            # comparison between ndarray and datetime64
            existing_axis < intersection.begin)[0]  # type: ignore
        # pylint: enable=comparison-with-callable
        return existing_ds.isel({
            partitioning_dim: indices
        }).concat(inserted_ds, partitioning_dim)

    assert relation.is_inside()
    # comparison between ndarray and datetime64
    index = numpy.where(existing_axis < intersection.begin)[0]  # type: ignore
    before: dataset.Dataset = existing_ds.isel(
        {partitioning_dim: slice(0, index[-1] + 1, None)})

    # pylint: disable=comparison-with-callable
    # comparison between ndarray and datetime64
    index = numpy.where(existing_axis > intersection.end())[0]  # type: ignore
    # pylint: enable=comparison-with-callable
    after: dataset.Dataset = existing_ds.isel(
        {partitioning_dim: slice(index[0], index[-1] + 1, None)})

    return before.concat((inserted_ds, after), partitioning_dim)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
I/O operations
==============
"""
from __future__ import annotations

from typing import Any, Iterable, Sequence
import collections
import json
import logging

import dask.array.core
import dask.base
from dask.delayed import Delayed as dask_Delayed
import dask.distributed
import dask.local
import fsspec
import numcodecs.abc
import numcodecs.blosc
import numpy
import zarr

from . import dataset, meta, sync
from .fs_utils import join_path
from .type_hints import ArrayLike

#: Name of the attribute storing the names of the dimensions of an array.
DIMENSIONS = '_ARRAY_DIMENSIONS'

#: Configuration file that describes the attributes of an array.
ZATTRS = '.zattrs'

#: Configuration file that describes the attributes of a group.
ZGROUP = '.zgroup'

#: Module logger.
_LOGGER: logging.Logger = logging.getLogger(__name__)

#: Disable multithreading in Blosc to avoid competing with Dask.
numcodecs.blosc.use_threads = False


def execute_transaction(
    client: dask.distributed.Client,
    synchronizer: sync.Sync,
    futures: Sequence[dask.distributed.Future | dask_Delayed],
    **kwargs: Any,
) -> Any:
    """Execute a transaction in the collection.

    This function executes a transaction in the collection by computing the
    given futures using the provided Dask client. The synchronizer instance is
    used to handle access to critical resources. Any additional keyword
    arguments are passed to the `dask.distributed.compute` function.

    Args:
        client: The Dask client.
        synchronizer: The instance handling access to critical resources.
        futures: Lazy tasks to be done.
        **kwargs: Keyword arguments to pass to `dask.distributed.compute`.

    Returns:
        The result of the transaction.
    """
    if not futures:
        return None
    awaitables: Iterable[Any] = []
    try:
        with synchronizer:
            awaitables = client.compute(futures,
                                        **kwargs)  # type: ignore[arg-type]
            return client.gather(awaitables)
    except:  # noqa: E722
        # Before throwing the exception, we wait until all future scheduled
        # ones finished.
        dask.distributed.wait(awaitables)
        dask.distributed.wait(futures)
        raise


def _to_zarr(array: dask.array.core.Array, mapper: fsspec.FSMap, path: str,
             **kwargs) -> None:
    """Write a Dask array to a Zarr dataset.

    Args:
        array: The Dask array to write.
        mapper: The file system mapper.
        path: The path to the Zarr dataset.
        **kwargs: Additional keyword arguments to pass to the `zarr.create`
            function.
    """
    chunks: list[tuple[int, ...]] = [chunk[0] for chunk in array.chunks]
    target: dask.array.core.Array = zarr.create(
        shape=array.shape,
        chunks=chunks,  # type: ignore[arg-type]
        dtype=array.dtype,
        store=mapper,
        path=path,
        overwrite=True,
        write_empty_chunks=False,
        **kwargs)
    dask.array.core.store(array,
                          target,
                          flush=True,
                          lock=False,
                          compute=True,
                          scheduler=dask.local.get_sync,
                          return_stored=False)


def write_zattrs(
    dirname: str,
    variable: meta.Variable | dataset.Variable,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Write the attributes of a variable to a Zarr dataset.

    Args:
        dirname: The storage directory of the Zarr dataset.
        variable: The variable to process.
        fs: The file system on which the Zarr dataset is stored.
    """
    attrs = collections.OrderedDict(item.get_config()
                                    for item in variable.attrs)
    attrs[DIMENSIONS] = variable.dimensions
    with fs.open(join_path(dirname, variable.name, ZATTRS),
                 mode='w') as stream:
        json.dump(attrs, stream, indent=2)  # type: ignore[arg-type]


def write_zarr_variable(
    args: tuple[str, dataset.Variable],
    dirname: str,
    fs: fsspec.AbstractFileSystem,
    *,
    block_size_limit: int | None = None,
    chunks: dict[str, int | str] | None = None,
) -> None:
    """Write a variable to a Zarr dataset.

    Args:
        args: The arguments to the function:
            - Name of the variable to write.
            - The variable to write.
        dirname: The target directory.
        fs: The file system on which the Zarr dataset is stored.
        block_size_limit: Maximum size (in bytes) of a block/chunk. Defaults
            to :data:`zcollection.meta.BLOCK_SIZE_LIMIT`.
        chunks: Chunk size for each dimension. Defaults to ``None`` (i.e. the
            default chunk size is used).
    """
    name: str
    variable: dataset.Variable
    kwargs: dict[str, tuple[numcodecs.abc.Codec, ...]]

    name, variable = args
    kwargs = {'filters': variable.filters}
    data: dask.array.core.Array = variable.array if isinstance(
        variable, dataset.DelayedArray) else dask.array.core.from_array(
            variable.array)

    # If the user has not specified a chunk size, we use the default one.
    # Otherwise, we use the user's choice.
    block_size_limit = block_size_limit or meta.BLOCK_SIZE_LIMIT
    var_chunks: dict[int, int | str] = {
        ix: -1
        for ix in range(variable.ndim)
    } if chunks is None else {
        ix: chunks.get(dim, -1)
        for ix, dim in enumerate(variable.dimensions)
    }
    data = data.rechunk(
        var_chunks,  # type: ignore[arg-type]
        block_size_limit=block_size_limit,
    )

    _to_zarr(array=data,
             mapper=fs.get_mapper(dirname),
             path=name,
             compressor=variable.compressor,
             fill_value=variable.fill_value,
             **kwargs)
    write_zattrs(dirname, variable, fs)


def _write_meta(
    zds: dataset.Dataset,
    dirname: str,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Write the metadata of a dataset to a Zarr dataset.

    Args:
        zds: The dataset to process.
        dirname: The storage directory of the Zarr dataset.
        fs: The file system on which the Zarr dataset is stored.
    """
    attrs = collections.OrderedDict(item.get_config() for item in zds.attrs)
    with fs.open(join_path(dirname, ZATTRS), mode='w') as stream:
        json.dump(attrs, stream, indent=2)  # type: ignore[arg-type]

    with fs.open(join_path(dirname, ZGROUP), mode='w') as stream:
        json.dump(
            {'zarr_format': 2},
            stream,  # type: ignore[arg-type]
            indent=2,
        )
    zarr.consolidate_metadata(fs.get_mapper(dirname))  # type: ignore[arg-type]
    fs.invalidate_cache(dirname)


def write_zarr_group(
    zds: dataset.Dataset,
    dirname: str,
    fs: fsspec.AbstractFileSystem,
    synchronizer: sync.Sync,
    *,
    distributed: bool = True,
) -> None:
    """Write a partition of a dataset to a Zarr group.

    Args:
        zds: The dataset partition to write.
        dirname: The name of the partition.
        fs: The file system that the partition is stored on.
        synchronizer: The instance handling access to critical resources.
        distributed: Whether to use Dask distributed to write the variables
            in parallel. Defaults to ``True``.

    Writes the variables of the given dataset partition to a Zarr group
    located at the specified directory on the given file system. If
    `distributed` is `True`, the variables are written in parallel using
    Dask distributed. Otherwise, the variables are written sequentially.

    The `synchronizer` argument is an instance of `sync.Sync` that handles
    access to critical resources, such as the Zarr group's metadata and
    attributes. This ensures that multiple processes or threads do not
    attempt to modify the same resource at the same time.
    """
    with synchronizer:
        if distributed:
            with dask.distributed.worker_client() as client:
                iterables: list[tuple[str, Any]] = [
                    (name, client.scatter(variable))
                    for name, variable in zds.variables.items()
                ]
                futures: list[dask.distributed.Future] = client.map(
                    write_zarr_variable,
                    iterables,
                    block_size_limit=zds.block_size_limit,
                    chunks=zds.chunks,
                    dirname=dirname,
                    fs=fs,
                )
                execute_transaction(
                    client,
                    sync.NoSync(),
                    futures,
                    workers=dask.distributed.get_worker().address)
        else:
            tuple(
                map(
                    lambda item: write_zarr_variable(
                        item,
                        dirname,
                        fs,
                        chunks=zds.chunks,
                        block_size_limit=zds.block_size_limit,
                    ), zds.variables.items()))
        _write_meta(zds, dirname, fs)


def open_zarr_array(
    array: zarr.Array,
    name: str,
    *,
    delayed: bool = True,
) -> dataset.Variable:
    """Open a Zarr array as a Dask array or a NumPy array.

    Args:
        array: The Zarr array to open.
        name: The name of the variable.
        delayed: Whether to load the variable lazily. If True, returns a Dask
            array. If False, returns a NumPy array. Defaults to True.

    Returns:
        The variable as a Dask array or a NumPy array.
    """
    if delayed:
        return dataset.DelayedArray.from_zarr(array, name, DIMENSIONS)
    return dataset.Array.from_zarr(array, name, DIMENSIONS)


def open_zarr_group(
    dirname,
    fs: fsspec.AbstractFileSystem,
    *,
    delayed: bool = True,
    selected_variables: Iterable[str] | None = None,
) -> dataset.Dataset:
    """Open a Zarr group stored in a partition.

    Args:
        dirname: The name of the partition.
        fs: The file system that the partition is stored on.
        delayed: Whether to load the variables lazily. Defaults to True.
        selected_variables: The list of variables to retain from the Zarr
            group. If None, all variables are selected. Defaults to None.

    Returns:
        The Zarr group stored in the partition, with the specified variables
        and attributes.
    """
    _LOGGER.debug('Opening Zarr group %r', dirname)
    store: zarr.Group = zarr.open_consolidated(  # type: ignore[arg-type]
        fs.get_mapper(dirname), mode='r')
    # Ignore unknown variables to retain.
    selected_variables = set(selected_variables) & set(
        store) if selected_variables is not None else set(store)
    variables: list[dataset.Variable] = [
        open_zarr_array(
            store[name],  # type: ignore[arg-type]
            name,
            delayed=delayed) for name in selected_variables
    ]

    return dataset.Dataset(
        variables=variables,
        attrs=tuple(dataset.Attribute(*item) for item in store.attrs.items()),
        delayed=delayed,
    )


def update_zarr_array(
    dirname: str,
    array: ArrayLike,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Update a Zarr array with new data.

    Args:
        dirname: The directory where the Zarr array is stored.
        array: The new data to write to the array.
        fs: The file system where the Zarr array is stored.

    Notes:
        This function updates the entire Zarr array with the new data. If the
        array is a Dask array, it must be computed before writing to the Zarr.
        If the array is a masked array and the Zarr array has a fill value, the
        masked values are filled with the fill value before writing to the Zarr
        array.
    """
    _LOGGER.debug('Updating Zarr array %r', dirname)
    store: zarr.Array = zarr.open_array(fs.get_mapper(dirname), mode='a')

    if isinstance(array, dask.array.core.Array):
        array = array.compute()

    if isinstance(array,
                  numpy.ma.MaskedArray) and store.fill_value is not None:
        array = array.filled(store.fill_value)

    store[:] = array

    # Invalidate any cached directory information.
    fs.invalidate_cache(dirname)


def del_zarr_array(
    dirname: str,
    name: str,
    fs: fsspec.AbstractFileSystem,
) -> None:
    """Delete a variable from a Zarr dataset.

    Args:
        dirname: The name of the dataset.
        name: The name of the variable.
        fs: The file system that the dataset is stored on.
    """
    _LOGGER.debug('Deleting Zarr array %r', dirname)
    path: str = join_path(dirname, name)
    if fs.exists(path):
        fs.rm(path, recursive=True)
        zarr.consolidate_metadata(
            fs.get_mapper(dirname),  # type: ignore[arg-type]
        )
        # Invalidate any cached directory information.
        fs.invalidate_cache(dirname)


def add_zarr_array(
    dirname: str,
    variable: meta.Variable,
    template: str,
    fs: fsspec.AbstractFileSystem,
    *,
    chunks: dict[str, int | str] | None = None,
) -> None:
    """Add a variable to a Zarr dataset.

    Args:
        dirname: The name of the dataset.
        variable: The variable to add.
        template: The name of the template variable.
        fs: The file system that the dataset is stored on.
        chunks: Chunk size for each dimension. Defaults to None. See
            :func:`zarr.create` for more information.

    Notes:
        This function adds a new variable to an existing Zarr dataset. The new
        variable is created with the same shape as the template variable, and
        with the specified chunk size (if provided). The function also writes
        the variable's attributes to the dataset, and consolidates the
        dataset's metadata.
    """
    _LOGGER.debug('Adding variable %r to Zarr dataset %r', variable.name,
                  dirname)
    shape: tuple[int, ...] = zarr.open(  # type: ignore[arg-type]
        fs.get_mapper(join_path(dirname, template))).shape

    var_chunks: tuple[int | str, ...] = shape if chunks is None else tuple(
        chunks.get(dim, shape[ix])  # type: ignore[misc]
        for ix, dim in enumerate(variable.dimensions))

    store: fsspec.FSMap = fs.get_mapper(join_path(dirname, variable.name))
    zarr.create(
        shape,
        chunks=var_chunks,  # type: ignore[arg-type]
        dtype=variable.dtype,
        compressor=variable.compressor,  # type: ignore[arg-type]
        fill_value=variable.fill_value,  # type: ignore[arg-type]
        store=store,
        filters=variable.filters)
    write_zattrs(dirname, variable, fs)
    zarr.consolidate_metadata(fs.get_mapper(dirname))  # type: ignore[arg-type]



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing variables
=================
"""
from typing import Any

import dask.array.core
import dask.array.ma
import numpy
import pytest

from ..array import _as_numpy_array


def test_as_numpy_array() -> None:
    """Test converting array like to a dask array."""
    array: numpy.ndarray
    fill_value: Any
    np_array: numpy.ndarray

    np_array = numpy.arange(10)
    array, fill_value = _as_numpy_array(np_array)
    assert isinstance(array, numpy.ndarray)
    assert not isinstance(array, numpy.ma.MaskedArray)
    assert fill_value is None

    np_array = numpy.ma.masked_equal(np_array, 5)
    array, fill_value = _as_numpy_array(np_array)
    assert isinstance(array, numpy.ndarray)
    assert not isinstance(array, numpy.ma.MaskedArray)
    assert fill_value == 5

    array, fill_value = _as_numpy_array(dask.array.ma.masked_equal(
        np_array, 5))
    assert isinstance(array, numpy.ndarray)
    assert not isinstance(array, numpy.ma.MaskedArray)
    assert fill_value == 5

    with pytest.raises(ValueError):
        _as_numpy_array(numpy.ma.masked_equal(np_array, 5), fill_value=6)

    with pytest.raises(ValueError):
        _as_numpy_array(numpy.ma.masked_equal(
            numpy.arange(numpy.datetime64(0, 'Y'),
                         numpy.datetime64(10, 'Y'),
                         dtype='M8[Y]'), numpy.datetime64(5, 'Y')),
                        fill_value=numpy.datetime64('NaT'))

    array, fill_value = _as_numpy_array(numpy.ma.masked_equal(
        numpy.arange(numpy.datetime64(0, 'Y'),
                     numpy.datetime64(10, 'Y'),
                     dtype='M8[Y]'), numpy.datetime64('NaT')),
                                        fill_value=numpy.datetime64('NaT'))
    assert isinstance(array, numpy.ndarray)
    assert not isinstance(array, numpy.ma.MaskedArray)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Callable objects.
=================
"""
from __future__ import annotations

from typing import Any, Callable, Protocol, Sequence

from .. import dataset
from ..type_hints import ArrayLike

#: Function type to load and call a callback function of type
#: :class:`PartitionCallable`.
WrappedPartitionCallable = Callable[[Sequence[str]], None]


#: pylint: disable=too-few-public-methods
class PartitionCallable(Protocol):
    """Protocol for partition callables.

    A partition callable is a function that accepts a dataset and
    returns a result.
    """

    @property
    def __name__(self) -> str:
        """Name of the callable."""
        # pylint: disable=unnecessary-ellipsis
        # Make checker happy.
        ...
        # pylint: enable=unnecessary-ellipsis

    def __call__(self, zds: dataset.Dataset, *args, **kwargs) -> Any:
        """Call the partition function.

        Args:
            zds: Dataset to partition.
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            Result of the partition function.
        """


#: Alias for :class:`PartitionCallable`.
MapCallable = PartitionCallable


class UpdateCallable(Protocol):
    """Protocol for update callables.

    A callable update is a function that accepts a data set and returns
    a dictionary of arrays to update.
    """

    @property
    def __name__(self) -> str:
        """Name of the callable."""
        # pylint: disable=unnecessary-ellipsis
        # Make checker happy.
        ...
        # pylint: enable=unnecessary-ellipsis

    def __call__(self, zds: dataset.Dataset, *args,
                 **kwargs) -> dict[str, ArrayLike]:
        """Call the update function.

        Args:
            zds: Dataset to update.
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            Dictionary of arrays to update.
        """
        # pylint: disable=unnecessary-ellipsis
        # Mandatory to make Pylance happy.
        ...
        # pylint: enable=unnecessary-ellipsis



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing utilities
=================
"""
import dask.distributed
import pytest

from .. import dask_utils
# pylint: disable=unused-import # Need to import for fixtures
from .cluster import dask_client, dask_cluster

# pylint: disable=unused-import


@pytest.mark.filterwarnings('ignore:Port \\d+ is already in use.*')
def test_get_client_with_no_cluster():
    """Test the get_client function with no cluster."""
    with dask_utils.get_client() as client:
        assert isinstance(client, dask.distributed.Client)


def test_get_client_with_cluster(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test the get_client function with a cluster."""
    with dask_utils.get_client() as client:
        assert isinstance(client, dask.distributed.Client)


def test_dask_workers(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
):
    """Test the dask_workers function."""
    assert dask_utils.dask_workers(dask_client, cores_only=True) == len(
        dask_client.ncores())  # type: ignore
    assert dask_utils.dask_workers(dask_client, cores_only=False) == sum(
        item for item in dask_client.nthreads().values())  # type: ignore


def test_split_sequence():
    """Test the split_sequence function."""
    assert list(dask_utils.split_sequence(list(range(10)), 2)) == [
        [0, 1, 2, 3, 4],
        [5, 6, 7, 8, 9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)), 3)) == [
        [0, 1, 2, 3],
        [4, 5, 6],
        [7, 8, 9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)), 4)) == [
        [0, 1, 2],
        [3, 4, 5],
        [6, 7],
        [8, 9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)), 5)) == [
        [0, 1],
        [2, 3],
        [4, 5],
        [6, 7],
        [8, 9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)), 6)) == [
        [0, 1],
        [2, 3],
        [4, 5],
        [6, 7],
        [8],
        [9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)), 7)) == [
        [0, 1],
        [2, 3],
        [4, 5],
        [6],
        [7],
        [8],
        [9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)), 8)) == [
        [0, 1],
        [2, 3],
        [4],
        [5],
        [6],
        [7],
        [8],
        [9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)), 9)) == [
        [0, 1],
        [2],
        [3],
        [4],
        [5],
        [6],
        [7],
        [8],
        [9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)), 10)) == [
        [0],
        [1],
        [2],
        [3],
        [4],
        [5],
        [6],
        [7],
        [8],
        [9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)), 11)) == [
        [0],
        [1],
        [2],
        [3],
        [4],
        [5],
        [6],
        [7],
        [8],
        [9],
    ]
    assert list(dask_utils.split_sequence(list(range(10)))) == [
        [0],
        [1],
        [2],
        [3],
        [4],
        [5],
        [6],
        [7],
        [8],
        [9],
    ]
    with pytest.raises(ValueError):
        list(dask_utils.split_sequence(list(range(10)), 0))



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Partitioning scheme.
====================
"""
from __future__ import annotations

from typing import (
    Any,
    Callable,
    ClassVar,
    Generator,
    Iterator,
    Match,
    Optional,
    Sequence,
    Tuple,
)
import abc
import collections
import re

import dask.array.core
import dask.array.creation
import dask.array.reductions
import dask.array.wrap
import fsspec
import numpy

from .. import dataset
from ..type_hints import ArrayLike, DTypeLike, NDArray

#: Object that represents a partitioning scheme
Partition = Tuple[Tuple[Tuple[str, Any], ...], slice]

#: The callable that parses the partitioning scheme
PatternType = Callable[[str], Optional[Match[str]]]

#: Allowed data types for partitioning schemes
DATA_TYPES = ('int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32',
              'uint64')


def _logical_or_reduce(
    arr: dask.array.core.Array,
    *,
    axis: int | tuple[int, ...] | None = None,
) -> dask.array.core.Array:
    """Implementation of `numpy.logical_or` reduction with dask.

    Args:
        arr: Array to reduce.
        axis: Axis to reduce. If this is None, a reduction is performed over
            all the axes. If this is a tuple of ints, a reduction is performed
            on multiple axes, instead of a single axis or all the axes as
            before.
    Returns:
        Reduced array.
    """
    axis = axis or 0

    #: pylint: disable=unused-argument
    # The function signature is required by the `dask.array.reduce` function.
    def chunk(block, axis, keepdims) -> Any:
        return block

    def aggregate(block, axis, keepdims) -> Any:
        return numpy.logical_or.reduce(block, axis=axis)

    #: pylint: enable=unused-argument

    return dask.array.reductions.reduction(
        arr[1:] != arr[:-1],  # type: ignore
        chunk=chunk,
        aggregate=aggregate,
        axis=axis,
        keepdims=False,
        dtype=numpy.bool_)


def unique(arr: dask.array.core.Array) -> tuple[NDArray, NDArray]:
    """Return unique elements and their indices.

    Args:
        arr: Array of elements.
    Returns:
        Tuple of unique elements and their indices.
    """
    size: int = arr.shape[0]
    chunks: tuple[int, ...] = arr.chunks[0]
    #: pylint: disable=not-callable
    mask: dask.array.core.Array = dask.array.wrap.empty((size, ),
                                                        dtype=numpy.bool_,
                                                        chunks=chunks)
    #: pylint: enable=not-callable
    mask[0] = True
    mask[1:] = (_logical_or_reduce(arr, axis=1)
                if arr.ndim > 1 else arr[1:] != arr[:-1])
    dtype: DTypeLike = numpy.uint32 if size < 2**32 else numpy.uint64
    indices: dask.array.core.Array = dask.array.creation.arange(
        size,
        dtype=dtype,
        chunks=chunks,  # type: ignore[arg-type]
    )
    mask = mask.persist()
    return arr[mask].compute(), indices[mask].compute()


def unique_and_check_monotony(arr: ArrayLike) -> tuple[NDArray, NDArray]:
    """Return unique elements and their indices.

    Args:
        arr: Array of elements.
        is_delayed: If True, the array is delayed.
    Returns:
        Tuple of unique elements and their indices.
    """
    index: NDArray
    indices: NDArray

    index, indices = numpy.unique(arr, axis=0, return_index=True)
    if not numpy.all(numpy.diff(indices) > 0):
        raise ValueError('index is not monotonic')
    return index, indices


def difference(arr: NDArray) -> NDArray:
    """Calculate the difference between each element in the array and the
    previous element.

    Args:
        arr: Array to calculate the difference for.
    Returns:
        Array of differences
    """
    return arr[1:] - arr[:-1]  # type: ignore


def concatenate_item(arr: NDArray, item: Any) -> NDArray:
    """Concatenate an array with a given item.

    Args:
        arr: Array to concatenate.
        item: Item to concatenate.
    Returns:
        Concatenated array.
    """
    return numpy.concatenate([arr, numpy.array([item], dtype=arr.dtype)])


def list_partitions(
    fs: fsspec.AbstractFileSystem,
    path: str,
    depth: int,
    root: bool = True,
) -> Iterator[str]:
    """The number of variables used for partitioning.

    The function will go down the tree and return all the files present when the
    requested depth is reached.
    Args:
        fs: file system object
        path: path to the directory
        depth: maximum depth of the directory tree.
        root: if True, the path is the root of the tree.
    Returns:
        Iterator of (path, directories, files).
    """
    if depth == -1:
        return StopIteration()

    if root:
        folders = map(
            lambda info: info['name'].rstrip('/'),
            filter(
                lambda info: info['type'] == 'directory' and not info['name'].
                split(fs.sep)[-1].startswith('.'),
                fs.ls(path, detail=True),
            ),
        )
        # If we're partitioning at top level (example: by year)
        if depth == 0:
            yield from sorted(folders)
            return StopIteration()

        for pathname in sorted(folders):
            yield from list_partitions(fs,
                                       pathname,
                                       depth=depth - 1,
                                       root=False)
        return StopIteration()

    if depth == 0:
        yield from sorted(fs.ls(path, detail=False))
        return StopIteration()

    for item in sorted(fs.ls(path, detail=False)):
        yield from list_partitions(fs, item, depth=depth - 1, root=False)
    return StopIteration()


class Partitioning(metaclass=abc.ABCMeta):
    """Initializes a new Partitioning instance.

    Args:
        variables: A list of strings representing the variables to be used for
            partitioning.
        dtype: An optional sequence of strings representing the data type used
            to store variable values in a binary representation without data
            loss. Must be one of the following allowed data types: ``int8``,
            ``int16``, ``int32``, ``int64``, ``uint8``, ``uint16``, ``uint32``,
            ``uint64``. If not provided, defaults to ``int64`` for all
            variables.

    Raises:
        TypeError: If dtype is not a sequence of strings.
        ValueError: If any of the data types provided is not one of the allowed
            data types.
    """
    __slots__ = ('_dtype', '_pattern', 'variables')

    #: The ID of the partitioning scheme
    ID: ClassVar[str | None] = None

    def __init__(self,
                 variables: Sequence[str],
                 dtype: Sequence[str] | None = None) -> None:
        if isinstance(dtype, str):
            raise TypeError('dtype must be a sequence of strings')
        if len(variables) == 0:
            raise ValueError('variables must not be empty')
        #: Variables to be used for the partitioning.
        self.variables = tuple(variables)
        #: Data type used to store variable values in a binary representation
        #: without data loss.
        self._dtype: tuple[str, ...] = tuple(
            dtype) if dtype is not None else ('int64', ) * len(self.variables)
        #: The regular expression that matches the partitioning scheme.
        self._pattern: PatternType = self._regex().search

        if len(set(self._dtype) - set(DATA_TYPES)) != 0:
            raise ValueError(
                f"Data type must be one of {', '.join(DATA_TYPES)}.")

    def __len__(self) -> int:
        """Return the number of partitions."""
        return len(self._dtype)

    def dtype(self) -> tuple[tuple[str, str], ...]:
        """Return the data type of the partitioning scheme."""
        return tuple(zip(self._keys(), self._dtype))

    def _keys(self) -> Sequence[str]:
        """Return the different keys of a partition."""
        return self.variables

    def _regex(self) -> re.Pattern:
        """Return a regular expression that matches the partitioning scheme."""
        return re.compile('.'.join(f'({item})=(.*)' for item in self._keys()))

    @abc.abstractmethod
    def _split(
        self,
        variables: dict[str, ArrayLike],
    ) -> Iterator[Partition]:
        """Split the variables constituting the partitioning into partitioning
        schemes.

        Args:
            variables: The variables to be split constituting the
                partitioning scheme.
        Returns:
            A sequence of tuples that contains the partitioning
                scheme and the associated indexer to divide the dataset on each
                partition found..
        """

    @staticmethod
    def _partition(selection: tuple[tuple[str, Any], ...]) -> tuple[str, ...]:
        """Format the partitioning scheme."""
        return tuple(f'{k}={v}' for k, v in selection)

    def index_dataset(
        self,
        zds: dataset.Dataset,
    ) -> Iterator[Partition]:
        """Yield the indexing scheme for the given dataset.

        Args:
            zds: The dataset to be indexed.
        Yields:
            The indexing scheme for the partitioning scheme.
        Raises:
            ValueError: if one of the variables needs for the partitioning
                is not monotonic.
        """
        variables = collections.OrderedDict(
            (name, zds.variables[name].array) for name in self.variables)
        if zds.delayed:
            # If the dask array is too chunked, the calculation is excessively
            # long.
            return self._split({
                name: arr.rechunk().persist()
                for name, arr in variables.items()
            })
        return self._split(variables)

    def split_dataset(
        self,
        zds: dataset.Dataset,
        axis: str,
    ) -> Iterator[tuple[tuple[str, ...], dict[str, slice]]]:
        """Split the dataset into partitions.

        Args:
            zds:  The dataset to be split.
            axis: The axis to be used for the splitting.
        Yields:
            The partitioning scheme and the indexer to divide the dataset on
            each partition found.
        Raises:
            ValueError: if one of the variables needs for the partitioning
                is not a one-dimensional array.
        """
        for item in self.variables:
            if len(zds.variables[item].shape) != 1:
                raise ValueError(f'f{item!r} must be a one-dimensional array')
        return ((self._partition(selection), {
            axis: indexer
        }) for selection, indexer in self.index_dataset(zds))

    def get_config(self) -> dict[str, Any]:
        """Return the configuration of the partitioning scheme.

        Returns:
            The configuration of the partitioning scheme.
        """
        config: dict[str, str | None] = {'id': self.ID}
        slots: Generator[tuple[str, ...], None, None] = (getattr(
            _class, '__slots__',
            ()) for _class in reversed(self.__class__.__mro__))
        config.update((attr, getattr(self, attr)) for _class in slots
                      for attr in _class if not attr.startswith('_'))
        return config

    @classmethod
    def from_config(cls, config) -> Partitioning:
        """Create a partitioning scheme from a configuration.

        Args:
            config: The configuration of the partitioning scheme.
        Returns:
            The partitioning scheme.
        """
        return cls(**config)

    def parse(self, partition: str) -> tuple[tuple[str, int], ...]:
        """Parse a partitioning scheme.

        Args:
            partition: The partitioning scheme to be parsed.
        Returns:
            The parsed partitioning scheme.
        Raises:
            ValueError: if the partitioning scheme is not driven by this
                instance.
        """
        match: Match[str] | None = self._pattern(partition)
        if match is None:
            raise ValueError(
                f'Partition is not driven by this instance: {partition}')
        groups: tuple[str, ...] = match.groups()
        return tuple((groups[ix], int(groups[ix + 1]))
                     for ix in range(0, len(groups), 2))

    @abc.abstractmethod
    def encode(
        self,
        partition: tuple[tuple[str, int], ...],
    ) -> tuple[Any, ...]:
        """Encode a partitioning scheme to the handled values.

        Args:
            partition: The partitioning scheme to be encoded.
        Returns:
            The encoded partitioning scheme.
        """

    @abc.abstractmethod
    def decode(self, values: tuple[Any, ...]) -> tuple[tuple[str, int], ...]:
        """Decode a partitioning scheme.

        Args:
            values: The encoded partitioning scheme.
        Returns:
            The decoded partitioning scheme.
        """

    @staticmethod
    def join(partition_scheme: tuple[tuple[str, int], ...], sep: str) -> str:
        """Join a partitioning scheme.

        Args:
            partition_scheme: The partitioning scheme to be joined.
            sep: The separator to be used.
        Returns:
            The joined partitioning scheme.
        """
        return sep.join(f'{k}={v}' for k, v in partition_scheme)

    def list_partitions(
        self,
        fs: fsspec.AbstractFileSystem,
        path: str,
    ) -> Iterator[str]:
        """List the partitions.

        Args:
            fs: The filesystem to be used.
            path: The path to the directory containing the partitions.
        Yields:
            The partitions.
        """
        return list_partitions(fs, path, depth=len(self) - 1)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Partitioning by date
====================
"""
from __future__ import annotations

from typing import Any, ClassVar, Iterator, Sequence
import datetime

import dask.array.core
import numpy

from . import abc
from ..type_hints import ArrayLike, NDArray

#: Numpy time units
RESOLUTION = ('Y', 'M', 'D', 'h', 'm', 's')

#: Numpy time unit meanings
UNITS = ('year', 'month', 'day', 'hour', 'minute', 'second')

#: Data type for time units
DATA_TYPES = ('uint16', 'uint8', 'uint8', 'uint8', 'uint8', 'uint8')

#: Time separation units
SEPARATORS: dict[str, str] = {
    'year': '-',
    'month': '-',
    'day': 'T',
    'hour': ':',
    'minute': ':',
    'second': '.'
}


def _unique(arr: ArrayLike, is_delayed: bool) -> tuple[NDArray, NDArray]:
    """Return unique elements and their indices.

    Args:
        arr: Array of elements.
        is_delayed: If True, the array is delayed.
    Returns:
        Tuple of unique elements and their indices.
    Raises:
        ValueError: If the array is not monotonic.
    """
    index: NDArray
    indices: NDArray

    if is_delayed:
        index, indices = abc.unique(arr)  # type: ignore[arg-type]
        # We don't use here the function `numpy.diff` but `abc.difference` for
        # optimization purposes.
        if not numpy.all(
                abc.difference(index.view(numpy.int64)) >= 0):  # type: ignore
            raise ValueError('index is not monotonic')
        return index, indices
    return abc.unique_and_check_monotony(arr)


class Date(abc.Partitioning):
    """Initialize a partitioning scheme based on dates.

    Args:
        variables: A list of strings representing the variables to be used for
            partitioning.
        resolution: Time resolution of the partitioning. Must be in
            :data:`RESOLUTION`.

    Raises:
        ValueError: If the resolution is not in the list of supported
            resolutions or if the partitioning is not performed on a one
            dimensional variable.

    Example:
        >>> partitioning = Date(variables=("time", ), resolution="Y")
    """
    __slots__ = ('_attrs', '_index', 'resolution')

    #: The ID of the partitioning scheme
    ID: ClassVar[str] = 'Date'

    def __init__(self, variables: Sequence[str], resolution: str) -> None:
        if len(variables) != 1:
            raise ValueError(
                'Partitioning on dates is performed on a single variable.')
        if resolution not in RESOLUTION:
            raise ValueError('resolution must be in: ' + ', '.join(RESOLUTION))
        index: int = RESOLUTION.index(resolution) + 1

        #: The time resolution of the partitioning
        self.resolution: str = resolution
        #: The time parts used for the partitioning
        self._attrs: tuple[str, ...] = UNITS[:index + 1]
        #: The indices of the time parts used for the partitioning
        self._index = tuple(range(index))
        super().__init__(variables,
                         tuple(DATA_TYPES[ix] for ix in self._index))

    def _keys(self) -> Sequence[str]:
        """Return the keys of the partitioning scheme."""
        return tuple(UNITS[ix] for ix in self._index)

    # pylint: disable=arguments-differ
    # False positive: the base method is static.
    def _partition(  # type: ignore[override]
        self,
        selection: tuple[tuple[str, Any], ...],
    ) -> tuple[str, ...]:
        """Return the partitioning scheme for the given selection."""
        datetime64: NDArray = selection[0][1]
        py_datetime: datetime.datetime = datetime64.astype('M8[s]').item()
        return tuple(UNITS[ix] + '=' +
                     f'{getattr(py_datetime, self._attrs[ix]):02d}'
                     for ix in self._index)
        # pylint: enable=arguments-differ

    def _split(
        self,
        variables: dict[str, ArrayLike],
    ) -> Iterator[abc.Partition]:
        """Return the partitioning scheme for the given variables."""
        index: NDArray
        indices: NDArray
        name: str
        values: ArrayLike

        # Determine if the variables are handled by Dask.
        is_delayed: bool = any(
            isinstance(value, dask.array.core.Array)
            for value in variables.values())
        name, values = tuple(variables.items())[0]

        if not numpy.issubdtype(values.dtype, numpy.dtype('datetime64')):
            raise TypeError('values must be a datetime64 array')

        index, indices = _unique(
            values.astype(f'datetime64[{self.resolution}]'), is_delayed)
        indices = abc.concatenate_item(indices, values.size)

        return ((((name, date), ), slice(start, indices[ix + 1], None))
                for date, (ix, start) in zip(index, enumerate(indices[:-1])))

    @staticmethod
    def _stringify(partition: tuple[tuple[str, int], ...]) -> str:
        """Return a string representation of the partitioning scheme."""
        string = ''.join(f'{value:02d}' + SEPARATORS[item]
                         for item, value in partition)
        if string[-1] in SEPARATORS.values():
            string = string[:-1]
        return string

    @staticmethod
    def join(partition_scheme: tuple[tuple[str, int], ...], sep: str) -> str:
        """Join a partitioning scheme.

        Args:
            partition_scheme: The partitioning scheme to be joined.
            sep: The separator to be used.

        Returns:
            The joined partitioning scheme.

        Example:
            >>> partitioning = Date(variables=("time", ), resolution="D")
            >>> partitioning.join((("year", 2020), ("month", 1), ("day", 1)),
            ...                   "/")
            'year=2020/month=01/day=01'
        """
        return sep.join(f'{k}={v:02d}' for k, v in partition_scheme)

    def encode(
        self,
        partition: tuple[tuple[str, int], ...],
    ) -> tuple[Any, ...]:
        """Encode a partitioning scheme.

        Args:
            partition: The partitioning scheme to be encoded.

        Returns:
            The encoded partitioning scheme.

        Example:
            >>> partitioning = Date(variables=("time", ), resolution="D")
            >>> fields = partitioning.parse("year=2020/month=01/day=01")
            >>> fields
            (("year", 2020), ("month", 1), ("day", 1))
            >>> partitioning.encode(fields)
            (numpy.datetime64('2020-01-01'),)
        """
        return tuple((numpy.datetime64(self._stringify(partition)), ))

    def decode(
        self,
        values: tuple[Any, ...],
    ) -> tuple[tuple[str, int], ...]:
        """Decode a partitioning scheme.

        Args:
            values: The partitioning scheme to be decoded.

        Returns:
            The decoded partitioning scheme.

        Example:
            >>> partitioning = Date(variables=("time", ), resolution="D")
            >>> partitioning.decode((numpy.datetime64('2020-01-01'), ))
            (("year", 2020), ("month", 1), ("day", 1))
        """
        datetime64: NDArray = values[0]
        py_datetime: datetime.datetime = datetime64.astype('M8[s]').item()
        return tuple((UNITS[ix], getattr(py_datetime, self._attrs[ix]))
                     for ix in self._index)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
View on a reference collection.
===============================
"""
from __future__ import annotations

from typing import Any, ClassVar, Iterable, Iterator, Sequence
import copy
import json
import logging
import pathlib
import warnings

import dask.array.core
import dask.bag.core
import dask.distributed
import dask.utils
import fsspec

from .. import collection, dask_utils, dataset, fs_utils, meta, storage, sync
from ..collection.callable_objects import MapCallable, PartitionCallable
from ..collection.detail import _try_infer_callable
from ..convenience import collection as convenience
from ..type_hints import ArrayLike
from .detail import (
    ViewReference,
    ViewUpdateCallable,
    _assert_have_variables,
    _assert_variable_handled,
    _create_zarr_array,
    _deserialize_filters,
    _drop_zarr_zarr,
    _load_datasets_list,
    _load_one_dataset,
    _select_overlap,
    _serialize_filters,
    _sync,
    _wrap_update_func,
    _wrap_update_func_overlap,
    _write_checksum,
)

__all__ = ('View', 'ViewReference')

#: Module logger.
_LOGGER = logging.getLogger(__name__)


class View:
    """View on a reference collection.

    Args:
        base_dir: Path to the directory where the view is stored.
        view_ref: Access properties for the reference view.
        ds: The dataset handled by this view.
        filesystem: The file system used to access the view.
        filters: The filters used to select the partitions of the reference. If
            not provided, all partitions are selected.
        synchronizer: The synchronizer used to synchronize the view.

    Note:
        Normally, you should not call this constructor directly. Instead, use
        :func:`create_view <zcollection.create_view>` or :func:`open_view
        <zcollection.open_view>` to create or open a view.
    """
    #: Configuration filename of the view.
    CONFIG: ClassVar[str] = '.view'

    def __init__(
        self,
        base_dir: str,
        view_ref: ViewReference,
        *,
        ds: meta.Dataset | None,
        filesystem: fsspec.AbstractFileSystem | str | None = None,
        filters: collection.PartitionFilter = None,
        synchronizer: sync.Sync | None = None,
    ) -> None:
        #: The file system used to access the view (default local file system).
        self.fs: fsspec.AbstractFileSystem = fs_utils.get_fs(filesystem)
        #: Path to the directory where the view is stored.
        self.base_dir: str = fs_utils.normalize_path(self.fs, base_dir)
        #: The reference collection of the view.
        self.view_ref: collection.Collection = convenience.open_collection(
            view_ref.path, mode='r', filesystem=view_ref.filesystem)
        #: The metadata of the variables handled by the view.
        self.metadata: meta.Dataset = ds or meta.Dataset(
            self.view_ref.metadata.dimensions, variables=[], attrs=[])
        #: The synchronizer used to synchronize the view.
        self.synchronizer: sync.Sync = synchronizer or sync.NoSync()
        #: The filters used to select the partitions of the reference.
        self.filters = filters

        if not self.fs.exists(self.base_dir):
            _LOGGER.info('Creating view %s', self)
            self.fs.makedirs(self.base_dir)
            self._write_config()
            self._init_partitions(filters)
        else:
            _LOGGER.info('Opening view %s', self)

    def _init_partitions(self, filters: collection.PartitionFilter) -> None:
        """Initialize the partitions of the view."""
        _LOGGER.info('Populating view %s', self)
        args = tuple(
            map(lambda item: fs_utils.join_path(self.base_dir, item),
                self.view_ref.partitions(filters=filters, relative=True)))
        # When opening an existing view, if the user asks to use new partitions
        # from the reference collection, in this case only the missing
        # partitions are created.
        args = tuple(filter(lambda item: not self.fs.exists(item), args))
        _LOGGER.info('%d partitions selected from %s', len(args),
                     self.view_ref)
        client: dask.distributed.Client = dask_utils.get_client()
        storage.execute_transaction(
            client, self.synchronizer,
            client.map(
                _write_checksum,
                tuple(args),
                base_dir=self.base_dir,
                view_ref=self.view_ref,
                fs=self.fs,
            ))

    def __str__(self) -> str:
        return (f'{self.__class__.__name__}'
                f'<filesystem={self.fs.__class__.__name__!r}, '
                f'base_dir={self.base_dir!r}>')

    @classmethod
    def _config(cls, base_dir: str) -> str:
        """Returns the configuration path."""
        return fs_utils.join_path(base_dir, cls.CONFIG)

    def _write_config(self) -> None:
        """Write the configuration file for the view."""
        config: str = self._config(self.base_dir)
        fs: dict[str, Any] = json.loads(self.view_ref.fs.to_json())
        with self.fs.open(config, mode='w') as stream:
            json.dump(
                {
                    'base_dir': self.base_dir,
                    'filters': _serialize_filters(self.filters),
                    'metadata': self.metadata.get_config(),
                    'view_ref': {
                        'path': self.view_ref.partition_properties.dir,
                        'fs': fs,
                    },
                },
                stream,  # type: ignore[arg-type]
                indent=4)
        self.fs.invalidate_cache(config)

    @classmethod
    def from_config(
        cls,
        path: str,
        *,
        filesystem: fsspec.AbstractFileSystem | str | None = None,
        synchronizer: sync.Sync | None = None,
    ) -> View:
        """Open a View described by a configuration file.

        Args:
            path: The path to the configuration file.
            filesystem: The filesystem to use for the view.
            synchronizer: The synchronizer to use for the view.

        Returns:
            The view.

        Raises:
            ValueError: If the provided directory does not contain a view.
        """
        _LOGGER.info('Opening view %r', path)
        fs: fsspec.AbstractFileSystem = fs_utils.get_fs(filesystem)
        config: str = cls._config(path)
        if not fs.exists(config):
            raise ValueError(f'zarr view not found at path {path!r}')
        with fs.open(config) as stream:
            data: dict[str, Any] = json.load(stream)

        view_ref: dict[str, Any] = data['view_ref']
        return View(data['base_dir'],
                    ViewReference(
                        view_ref['path'],
                        fsspec.AbstractFileSystem.from_json(
                            json.dumps(view_ref['fs']))),
                    ds=meta.Dataset.from_config(data['metadata']),
                    filesystem=filesystem,
                    filters=_deserialize_filters(data['filters']),
                    synchronizer=synchronizer)

    def partitions(
        self,
        filters: collection.PartitionFilter = None,
    ) -> Iterator[str]:
        """Returns the list of partitions in the view.

        Args:
            filters: The partition filters.

        Returns:
            The list of partitions.
        """
        return filter(
            self.fs.exists,
            map(lambda item: fs_utils.join_path(self.base_dir, item),
                self.view_ref.partitions(filters=filters, relative=True)))

    def variables(
        self,
        selected_variables: Iterable[str] | None = None
    ) -> tuple[dataset.Variable, ...]:
        """Return the variables of the view.

        Args:
            selected_variables: The variables to return. If None, all the
                variables are returned.

        Returns:
            The variables of the view.
        """
        return dataset.get_dataset_variable_properties(self.metadata,
                                                       selected_variables)

    def add_variable(
        self,
        variable: meta.Variable | dataset.Variable,
    ) -> None:
        """Add a variable to the view.

        Args:
            variable: The variable to add

        Raises:
            ValueError: If the variable already exists

        Example:
            >>> view.add_variable(
            ...    zcollection.meta.Variable(
            ...        "temperature",
            ...        "float32", ("time", "lat", "lon"),
            ...        (zcollection.meta.Attribute("units", "degrees Celsius"),
            ...         zcollection.meta.Attribute("long_name", "temperature")),
            ...        fill_value=-9999.0))
        """
        # pylint: disable=duplicate-code
        # false positive, no code duplication

        variable = dataset.get_variable_metadata(variable)
        _LOGGER.info('Adding variable %r in the view', variable.name)
        if (variable.name in self.view_ref.metadata.variables
                or variable.name in self.metadata.variables):
            raise ValueError(f'Variable {variable.name} already exists')
        client: dask.distributed.Client = dask_utils.get_client()
        self.metadata.add_variable(variable)
        template: meta.Variable = \
            self.view_ref.metadata.search_same_dimensions_as(variable)

        existing_partitions: set[str] = {
            pathlib.Path(path).relative_to(self.base_dir).as_posix()
            for path in self.partitions()
        }

        if len(existing_partitions) == 0:
            _LOGGER.info('No partitions found, skipping variable creation')
            return

        args = filter(lambda item: item[0] in existing_partitions,
                      self.view_ref.iterate_on_records(relative=True))

        # Remove the attribute from the variable. The attribute will be added
        # from the view metadata.
        # Remove the attribute from the variable. The attribute will be added
        # from the collection metadata.
        variable = variable.set_for_insertion()

        try:
            storage.execute_transaction(
                client, self.synchronizer,
                client.map(_create_zarr_array,
                           tuple(args),
                           base_dir=self.base_dir,
                           fs=self.fs,
                           template=template.name,
                           variable=variable))
        except Exception:
            storage.execute_transaction(
                client, self.synchronizer,
                client.map(_drop_zarr_zarr,
                           tuple(self.partitions()),
                           fs=self.fs,
                           variable=variable.name,
                           ignore_errors=True))
            raise

        self._write_config()
        # pylint: enable=duplicate-code

    def drop_variable(
        self,
        varname: str,
    ) -> None:
        """Drop a variable from the view.

        Args:
            varname: The name of the variable to drop.

        Raise:
            ValueError: If the variable does not exist or if the variable
                belongs to the reference collection.

        Example:
            >>> view.drop_variable("temperature")
        """
        _LOGGER.info('Dropping variable %r', varname)
        _assert_variable_handled(self.view_ref.metadata, self.metadata,
                                 varname)
        client: dask.distributed.Client = dask_utils.get_client()

        variable: meta.Variable = self.metadata.variables.pop(varname)
        self._write_config()

        storage.execute_transaction(
            client, self.synchronizer,
            client.map(_drop_zarr_zarr,
                       tuple(self.partitions()),
                       fs=self.fs,
                       variable=variable.name))

    def load(
        self,
        *,
        delayed: bool = True,
        filters: collection.PartitionFilter = None,
        indexer: collection.abc.Indexer | None = None,
        selected_variables: Iterable[str] | None = None,
    ) -> dataset.Dataset | None:
        """Load the view.

        Args:
            delayed: Whether to load data in a dask array or not.
            filters: The predicate used to filter the partitions to select.
                To get more information on the predicate, see the
                documentation of the :meth:`Collection.partitions
                <zcollection.collection.Collection.partitions>` method.
            indexer: The indexer to apply.
            selected_variables: A list of variables to retain from the view.
                If None, all variables are loaded.

        Returns:
            The dataset.

        Example:
            >>> view.load()
            >>> view.load(filters="time == '2020-01-01'")
            >>> view.load(filters=lambda x: x["time"] == "2020-01-01")
        """
        _assert_have_variables(self.metadata)
        if indexer is not None:
            arguments = tuple(
                collection.abc.build_indexer_args(
                    self.view_ref,
                    filters,
                    indexer,
                    partitions=self.partitions()))
            if len(arguments) == 0:
                return None
        else:
            arguments = tuple((self.view_ref.partitioning.parse(item), [])
                              for item in self.partitions(filters=filters))

        client: dask.distributed.Client = dask_utils.get_client()
        futures: list[dask.distributed.Future] = client.map(
            _load_one_dataset,
            arguments,
            base_dir=self.base_dir,
            delayed=delayed,
            fs=self.fs,
            selected_variables=self.view_ref.metadata.select_variables(
                selected_variables),
            view_ref=client.scatter(self.view_ref),
            variables=self.metadata.select_variables(selected_variables))

        # The load function returns the path to the partitions and the loaded
        # datasets. Only the loaded datasets are retrieved here and filter None
        # values corresponding to empty partitions.
        arrays: list[dataset.Dataset] = list(
            map(
                lambda item: item[0],  # type: ignore[arg-type]
                filter(lambda item: item is not None,
                       client.gather(futures))))  # type: ignore[arg-type]
        if arrays:
            array: dataset.Dataset = arrays.pop(0)
            if arrays:
                array = array.concat(arrays,
                                     self.view_ref.partition_properties.dim)
            metadata: meta.Dataset = copy.deepcopy(self.view_ref.metadata)
            metadata.variables.update(self.metadata.variables.items())
            array.fill_attrs(metadata)
            return array
        return None

    def update(
        self,
        func: collection.UpdateCallable,
        /,
        *args,
        depth: int = 0,
        delayed: bool = True,
        filters: collection.PartitionFilter = None,
        npartitions: int | None = None,
        selected_variables: Iterable[str] | None = None,
        trim: bool = True,
        **kwargs,
    ) -> None:
        """Update a variable stored int the view.

        Args:
            func: The function to apply to calculate the new values for the
                target variables.
            depth: The depth of the overlap between the partitions. Default is
                0 (no overlap). If depth is greater than 0, the function is
                applied on the partition and its neighbors selected by the
                depth. If ``func`` accepts a partition_info as a keyword
                argument, it will be passed a tuple with the name of the
                partitioned dimension and the slice allowing getting in the
                dataset the selected partition without the overlap.
            delayed: Whether to load data in a dask array or in memory.
            filters: The predicate used to filter the partitions to drop.
                To get more information on the predicate, see the
                documentation of the :meth:`Collection.partitions
                <zcollection.collection.Collection.partitions>` method.
            npartitions: The number of partitions to update in parallel. By
                default, it is equal to the number of Dask workers available
                when calling this method.
            selected_variables: A list of variables to retain from the view.
                If None, all variables are loaded. Useful to load only a
                subset of the view.
            trim: Whether to trim ``depth`` items from each partition after
                calling ``func``. Set it to ``False`` if your function does
                this for you.
            args: The positional arguments to pass to the function.
            kwargs: The keyword arguments to pass to the function.

        Raises:
            ValueError: If the variable does not exist or if the variable
                belongs to the reference collection.
            ValueError: If the depth is greater than 0 and the selected
                variables does not contain the variables updated by the
                function.

        Example:
            >>> def temp_celsius_to_kelvin(
            ...     dataset: zcollection.dataset.Dataset,
            ... ) -> Dict[str, numpy.ndarray]:
            ...     return dict(
            ...         temperature_kelvin=dataset["temperature"].values + 273,
            ...         15)
            >>> view.update(update_temperature)
        """
        _assert_have_variables(self.metadata)

        client: dask.distributed.Client = dask_utils.get_client()

        datasets_list = tuple(
            _load_datasets_list(client=client,
                                base_dir=self.base_dir,
                                delayed=delayed,
                                fs=self.fs,
                                view_ref=self.view_ref,
                                metadata=self.metadata,
                                partitions=self.partitions(filters),
                                selected_variables=selected_variables))

        # If no dataset is selected, we have nothing to do.
        if not datasets_list:
            warnings.warn('The update function is not applied because no '
                          'data is selected with the given filters.')
            return

        func_result: dict[str, ArrayLike] = _try_infer_callable(
            func, datasets_list[0][0], self.view_ref.partition_properties.dim,
            *args, **kwargs)
        tuple(
            map(
                lambda varname: _assert_variable_handled(
                    self.view_ref.metadata, self.metadata, varname),
                func_result))
        _LOGGER.info('Updating variable %s',
                     ', '.join(repr(item) for item in func_result))

        # Function to apply to each partition.
        wrap_function: ViewUpdateCallable

        # Wrap the function to apply to each partition.
        if depth == 0:
            wrap_function = _wrap_update_func(
                func,
                self.fs,
                *args,
                **kwargs,
            )
        else:
            if selected_variables is not None and len(
                    set(func_result) & set(selected_variables)) == 0:
                raise ValueError(
                    'If the depth is greater than 0, the selected variables '
                    'must contain the variables updated by the function.')

            wrap_function = _wrap_update_func_overlap(
                datasets_list,
                depth,
                func,
                self.fs,
                self.view_ref,
                trim,
                *args,
                **kwargs,
            )

        batchs: Iterator[Sequence[Any]] = dask_utils.split_sequence(
            datasets_list, npartitions
            or dask_utils.dask_workers(client, cores_only=True))
        awaitables: list[dask.distributed.Future] = client.map(
            wrap_function,
            tuple(batchs),
            key=func.__name__,
            base_dir=self.base_dir)
        storage.execute_transaction(client, self.synchronizer, awaitables)

    # pylint: disable=duplicate-code
    # false positive, no code duplication
    def map(
        self,
        func: MapCallable,
        /,
        *args,
        delayed: bool = True,
        filters: collection.PartitionFilter = None,
        partition_size: int | None = None,
        npartitions: int | None = None,
        selected_variables: Iterable[str] | None = None,
        **kwargs,
    ) -> dask.bag.core.Bag:
        """Map a function over the partitions of the view.

        Args:
            func: The function to apply to every partition of the view.
            *args: The positional arguments to pass to the function.
            delayed: Whether to load data in a dask array or in memory.
            filters: The predicate used to filter the partitions to process.
                To get more information on the predicate, see the
                documentation of the :meth:`zcollection.Collection.partitions`
                method.
            partition_size: The length of each bag partition.
            npartitions: The number of desired bag partitions.
            selected_variables: A list of variables to retain from the view.
                If None, all variables are loaded. Useful to load only a
                subset of the view.
            **kwargs: The keyword arguments to pass to the function.

        Returns:
            A bag containing the tuple of the partition scheme and the result
            of the function.

        Example:
            >>> futures = view.map(
            ...     lambda x: (x["var1"] + x["var2"]).values)
            >>> for item in futures:
            ...     print(item)
            [1.0, 2.0, 3.0, 4.0]
            [5.0, 6.0, 7.0, 8.0]
        """

        def _wrap(
            arguments: tuple[dataset.Dataset, str],
            func: PartitionCallable,
            *args,
            **kwargs,
        ) -> tuple[tuple[tuple[str, int], ...], Any]:
            """Wraps the function to apply on the partition.

            Args:
                arguments: The partition scheme and the dataset.
                func: The function to apply.
                *args: The positional arguments to pass to the function.
                **kwargs: The keyword arguments to pass to the function.

            Returns:
                The result of the function.
            """
            zds: dataset.Dataset
            partition: str

            zds, partition = arguments
            return self.view_ref.partitioning.parse(partition), func(
                zds, *args, **kwargs)

        _assert_have_variables(self.metadata)

        client: dask.distributed.Client = dask_utils.get_client()
        datasets_list = tuple(
            _load_datasets_list(client=client,
                                base_dir=self.base_dir,
                                delayed=delayed,
                                fs=self.fs,
                                view_ref=self.view_ref,
                                metadata=self.metadata,
                                partitions=self.partitions(filters),
                                selected_variables=selected_variables))
        bag: dask.bag.core.Bag = dask.bag.core.from_sequence(
            datasets_list,
            partition_size=partition_size,
            npartitions=npartitions)
        return bag.map(_wrap, func, *args, **kwargs)
        # pylint: enable=duplicate-code

    def map_overlap(
        self,
        func: MapCallable,
        /,
        *args,
        depth: int = 1,
        delayed: bool = True,
        filters: collection.PartitionFilter = None,
        partition_size: int | None = None,
        npartitions: int | None = None,
        selected_variables: Iterable[str] | None = None,
        **kwargs,
    ) -> dask.bag.core.Bag:
        """Map a function over the partitions of the view with some overlap.

        Args:
            func: The function to apply to every partition of the view.
            depth: The depth of the overlap between the partitions. Default is
                0 (no overlap). If depth is greater than 0, the function is
                applied on the partition and its neighbors selected by the
                depth. If ``func`` accepts a partition_info as a keyword
                argument, it will be passed a tuple with the name of the
                partitioned dimension and the slice allowing getting in the
                dataset the selected partition without the overlap.
            *args: The positional arguments to pass to the function.
            delayed: Whether to load data in a dask array or in memory.
            filters: The predicate used to filter the partitions to process.
                To get more information on the predicate, see the
                documentation of the :meth:`zcollection.Collection.partitions`
                method.
            partition_size: The length of each bag partition.
            npartitions: The number of desired bag partitions.
            selected_variables: A list of variables to retain from the view.
                If None, all variables are loaded. Useful to load only a
                subset of the view.
            **kwargs: The keyword arguments to pass to the function.

        Returns:
            A bag containing the tuple of the partition scheme and the result
            of the function.

        Example:
            >>> futures = view.map_overlap(
            ...     lambda x: (x["var1"] + x["var2"]).values,
            ...     depth=1)
            >>> for item in futures:
            ...     print(item)
            [1.0, 2.0, 3.0, 4.0]
        """
        if depth < 0:
            raise ValueError('Depth must be greater than or equal to 0')

        add_partition_info: bool = dask.utils.has_keyword(
            func, 'partition_info')

        def _wrap(
            arguments: tuple[dataset.Dataset, str],
            func: PartitionCallable,
            datasets_list: tuple[tuple[dataset.Dataset, str]],
            depth: int,
            *args,
            **kwargs,
        ) -> tuple[tuple[tuple[str, int], ...], Any]:
            """Wraps the function to apply on the partition.

            Args:
                arguments: The partition scheme and the dataset.
                func: The function to apply.
                datasets: The datasets to apply the function on.
                depth: The depth of the overlap between the partitions.
                *args: The positional arguments to pass to the function.
                **kwargs: The keyword arguments to pass to the function.

            Returns:
                The result of the function.
            """
            zds: dataset.Dataset
            indices: slice

            zds, indices = _select_overlap(arguments, datasets_list, depth,
                                           self.view_ref)

            if add_partition_info:
                kwargs = kwargs.copy()
                kwargs['partition_info'] = (
                    self.view_ref.partition_properties.dim, indices)

            # Finally, apply the function.
            return (self.view_ref.partitioning.parse(arguments[1]),
                    func(zds, *args, **kwargs))

        _assert_have_variables(self.metadata)

        client: dask.distributed.Client = dask_utils.get_client()
        datasets_list = tuple(
            _load_datasets_list(client=client,
                                base_dir=self.base_dir,
                                delayed=delayed,
                                fs=self.fs,
                                view_ref=self.view_ref,
                                metadata=self.metadata,
                                partitions=self.partitions(filters),
                                selected_variables=selected_variables))
        bag: dask.bag.core.Bag = dask.bag.core.from_sequence(
            datasets_list,
            partition_size=partition_size,
            npartitions=npartitions)
        return bag.map(_wrap, func, datasets_list, depth, *args, **kwargs)

    def is_synced(self) -> bool:
        """Check if the view is synchronized with the underlying collection.

        Returns:
            True if the view is synchronized, False otherwise.
        """
        partitions = tuple(self.view_ref.partitions(relative=True))
        client: dask.distributed.Client = dask_utils.get_client()
        unsynchronized_partition = storage.execute_transaction(
            client, self.synchronizer,
            client.map(_sync,
                       partitions,
                       base_dir=self.base_dir,
                       fs=self.fs,
                       view_ref=self.view_ref,
                       metadata=self.metadata,
                       dry_run=True))
        return len(
            tuple(
                filter(lambda item: item is not None,
                       unsynchronized_partition))) == 0

    def sync(
        self,
        filters: collection.PartitionFilter = None
    ) -> collection.abc.PartitionFilterCallback:
        """Synchronize the view with the underlying collection.

        This method is useful to update the view after a change in the
        underlying collection.

        Args:
            filters: The predicate used to select the partitions to
                synchronize. To get more information on the predicate, see the
                documentation of the :meth:`zcollection.Collection.partitions`
                method.
                If None, the view is synchronized with all the partitions
                already present in the view. If you want to extend the view
                with new partitions, use must provide a predicate that
                selects the new partitions.
                Existing partitions are not removed, even if they are not
                selected by the predicate.

        Returns:
            A function that can be used as a predicate to get the partitions
            that have been synchronized using the :meth:`View.partitions`
            method.
        """
        _LOGGER.info('Synchronizing view %s', self)

        if filters is not None:
            self.filters = filters
            self._write_config()
            self._init_partitions(filters)

        partitions = tuple(self.view_ref.partitions(relative=True))
        _LOGGER.info('%d partitions to synchronize', len(partitions))

        client: dask.distributed.Client = dask_utils.get_client()
        synchronized_partition: list[str | None] = storage.execute_transaction(
            client, self.synchronizer,
            client.map(_sync,
                       partitions,
                       base_dir=self.base_dir,
                       fs=self.fs,
                       view_ref=self.view_ref,
                       metadata=self.metadata))

        partition_ids = tuple(
            dict(self.view_ref.partitioning.parse(
                item))  # type: ignore[arg-type] # item is filtered
            for item in filter(lambda item: item is not None,
                               synchronized_partition))
        return lambda item: item in partition_ids



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing variables
=================
"""
from typing import Any
import pickle

import dask.array.core
import dask.array.ma
import numpy
import pytest
import xarray
import zarr

from ... import meta
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster
from ..abc import Variable
from ..array import Array
from ..delayed_array import DelayedArray
from .data import array, delayed_array

# pylint enable=unused-import

#: Create an Array for testing
ARRAY = array

#: Create a DelayedArray for testing
DELAYED_ARRAY = delayed_array


@pytest.mark.parametrize('factory', [ARRAY, DELAYED_ARRAY])
def test_masked_array(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test masked array."""
    var1: Variable = factory()
    var2: Variable = var1.rename('var2')
    assert var2.array is var1.array
    assert var2.name == 'var2'
    assert var2.dimensions == var1.dimensions
    assert var2.attrs == var1.attrs
    assert var2.compressor == var1.compressor
    assert var2.filters == var1.filters
    assert var2.fill_value == var1.fill_value
    assert var2.dtype == var1.dtype
    assert var2.shape == var1.shape
    assert var2.size == var1.size
    assert var2.ndim == var1.ndim


@pytest.mark.parametrize('factory', [ARRAY, DELAYED_ARRAY])
def test_constructor(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test variable creation."""
    var1: Variable = factory()
    assert var1.name == 'var1'
    assert var1.dtype == numpy.dtype('int64')
    assert var1.shape == (5, 2)
    assert var1.dimensions == ('x', 'y')
    assert var1.attrs == (meta.Attribute(name='attr', value=1), )
    assert var1.compressor.cname == 'zstd'  # type: ignore
    assert var1.compressor.clevel == 1  # type: ignore
    assert var1.fill_value == 0
    assert var1.size == 10
    assert var1.nbytes == 80
    assert var1.filters == (
        zarr.Delta('int64', 'int32'),
        zarr.Delta('int32', 'int32'),
    )
    assert numpy.all(var1.values == numpy.arange(10).reshape(5, 2))
    assert numpy.all(var1.values == var1.values)
    assert tuple(var1.dimension_index()) == (('x', 0), ('y', 1))
    assert isinstance(var1.metadata(), meta.Variable)
    assert isinstance(str(var1), str)
    assert isinstance(repr(var1), str)

    var2: Variable = pickle.loads(pickle.dumps(var1))
    assert var2.name == 'var1'
    assert var2.dtype == numpy.dtype('int64')
    assert var2.shape == (5, 2)
    assert var2.dimensions == ('x', 'y')
    assert var2.attrs == (meta.Attribute(name='attr', value=1), )
    assert var2.compressor.cname == 'zstd'  # type: ignore
    assert var2.compressor.clevel == 1  # type: ignore
    assert var2.fill_value == 0
    assert var2.size == 10
    assert var2.nbytes == 80
    assert var2.filters == (
        zarr.Delta('int64', 'int32'),
        zarr.Delta('int32', 'int32'),
    )
    assert numpy.all(var1.values == var2.values)

    def add_arrays(arr1, arr2) -> Any:
        return arr1 + arr2

    assert numpy.all(
        add_arrays(var1, var1.values) == numpy.arange(10).reshape(5, 2) +
        numpy.arange(10).reshape(5, 2))

    assert numpy.all(
        add_arrays(var1, var1.data).compute() ==
        numpy.arange(10).reshape(5, 2) + numpy.arange(10).reshape(5, 2))

    var1.values = numpy.ones((10, 4), dtype='int64')
    assert var1.data.shape == (10, 4)
    assert isinstance(var1.data, dask.array.core.Array)
    assert numpy.all(var1.values == 1)

    with pytest.raises(ValueError):
        var1.values = numpy.ones((10, 4, 2), dtype='int64')


@pytest.mark.parametrize('factory', [ARRAY, DELAYED_ARRAY])
def test_duplicate(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test of the duplication of variables."""
    var1: Variable = factory()
    var2: Variable = var1.duplicate(var1.array * 2)
    assert var2.name == 'var1'
    assert var2.dtype == numpy.dtype('int64')
    assert var2.shape == (5, 2)
    assert var2.dimensions == ('x', 'y')
    assert var2.attrs == (meta.Attribute(name='attr', value=1), )
    assert var2.compressor.cname == 'zstd'  # type: ignore
    assert var2.compressor.clevel == 1  # type: ignore
    assert var2.fill_value == 0
    assert var2.filters == (
        zarr.Delta('int64', 'int32'),
        zarr.Delta('int32', 'int32'),
    )
    assert numpy.all(var1.values == var2.values / 2)  # type: ignore
    assert var1.metadata() == var2.metadata()

    with pytest.raises(ValueError):
        var1.duplicate(numpy.ones((10, 4, 2), dtype='int64'))


@pytest.mark.parametrize('factory', [ARRAY, DELAYED_ARRAY])
def test_concatenate(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test concatenation of variables."""
    var1: Variable = factory()
    var2: Variable = factory()
    var3: Variable = factory()

    var4: Variable = var1.concat((var2, var3), 'x')
    assert numpy.all(var4.values == numpy.concatenate(
        (var1.values, var2.values, var3.values), axis=0))

    var4 = var1.concat(var2, 'x')
    assert numpy.all(
        var4.values == numpy.concatenate((var1.values, var2.values), axis=0))

    with pytest.raises(ValueError):
        var1.concat([], 'y')


@pytest.mark.parametrize('factory', [ARRAY, DELAYED_ARRAY])
def test_getitem(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test getting of variables."""
    var: Variable = factory()
    values: numpy.ndarray = var.values
    result: numpy.ndarray = var[0]
    assert numpy.all(result == values[0])
    result = var[0:2]
    assert numpy.all(result == values[0:2])
    result = var[0:2, 0]
    assert numpy.all(result == values[0:2, 0])
    result = var[0:2, 0:2]
    assert numpy.all(result == values[0:2, 0:2])


@pytest.mark.parametrize('factory', [ARRAY, DELAYED_ARRAY])
def test_fill(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test filling of variables."""
    var: Variable = factory()
    assert not var.values.all() is numpy.ma.masked
    var.fill()
    assert var.values.all() is numpy.ma.masked


@pytest.mark.parametrize('factory', [ARRAY, DELAYED_ARRAY])
def test_rechunk(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test rechunking of variables."""
    var: Variable = factory()
    values: numpy.ndarray = var.values
    var = var.rechunk()
    assert numpy.all(var.values == values)


@pytest.mark.parametrize('factory', [Array, DelayedArray])
def test_dimension_less(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Concatenate two dimensionless variables."""
    data: numpy.ndarray = numpy.array([0, 1], dtype=numpy.int32)
    var1: Variable = factory('nv',
                             data, ('nv', ),
                             attrs=(meta.Attribute('comment', 'vertex'),
                                    meta.Attribute('units', '1')))
    assert var1.fill_value is None
    metadata: meta.Variable = var1.metadata()
    assert metadata.fill_value is None
    assert meta.Variable.from_config(metadata.get_config()) == metadata

    var2: Variable = factory('nv',
                             data, ('nv', ),
                             attrs=(meta.Attribute('comment', 'vertex'),
                                    meta.Attribute('units', '1')))

    concatenated: Variable = var1.concat((var2, ), 'time')
    assert numpy.all(concatenated.values == var1.values)
    assert concatenated.metadata() == var1.metadata()


@pytest.mark.parametrize('factory', [Array, DelayedArray])
def test_timedelta64_to_xarray(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test conversion to xarray."""
    delta: numpy.ndarray = numpy.diff(
        numpy.arange(
            numpy.datetime64('2000-01-01', 'ns'),
            numpy.datetime64('2000-02-01', 'ns'),
            numpy.timedelta64('1', 'h'),
        ))

    var: Variable = factory(
        name='timedelta',
        data=delta,
        dimensions=('num_lines', ),
        attrs=(meta.Attribute(name='attr', value=1), ),
        compressor=zarr.Blosc(),
        filters=(zarr.Delta('int64', 'int64'), ),
    )
    xr_var: xarray.Variable = var.to_xarray()
    assert xr_var.dims == ('num_lines', )
    assert xr_var.attrs == {'attr': 1}
    assert xr_var.dtype.kind == 'm'


@pytest.mark.parametrize('factory', [Array, DelayedArray])
def test_datetime64_to_xarray(
        factory,
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test conversion to xarray."""
    dates: numpy.ndarray = numpy.arange(
        numpy.datetime64('2000-01-01', 'ns'),
        numpy.datetime64('2000-02-01', 'ns'),
        numpy.timedelta64('1', 'h'),
    )
    var: Variable = factory(
        name='time',
        data=dates,
        dimensions=('num_lines', ),
        attrs=(meta.Attribute(name='attr', value=1), ),
        compressor=zarr.Blosc(),
        filters=(zarr.Delta('int64', 'int64'), ),
    )
    xr_var: xarray.Variable = var.to_xarray()
    assert xr_var.dims == ('num_lines', )
    assert xr_var.attrs == {'attr': 1}
    assert xr_var.dtype == 'datetime64[ns]'



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Fixtures for testing S3 using the pytest and minio.
===================================================
"""
from typing import Iterator, Literal
import os
import pathlib
import shlex
import subprocess
import time

import botocore.client
import botocore.session
import pytest
import requests
import s3fs

#: Listen port
PORT = 5555
#: Listen address
ENDPOINT: str = f'127.0.0.1:{PORT}'
#: URI for minio
ENDPOINT_URI: str = f'http://{ENDPOINT}'
#: Credential for minio
CREDENTIAL = '25219d58-f6c6-11eb-922c-770d49cd18e4'


def have_minio() -> Literal[True]:
    """Check if minio is available."""
    try:
        subprocess.check_output(['minio', '--version'])
        return True
    except:
        raise ImportError('minio: command not found') from None


have_minio()


def is_minio_up(timeout: float) -> bool:
    """Check if minio server is up."""
    try:
        response = requests.get(ENDPOINT_URI, timeout=timeout)
        if response.status_code == 403:
            return True
    except:  # pylint: disable=bare-except
        pass
    return False


def wait_for_minio_to_start(timeout: float) -> None:
    """Wait for the minio server to start."""
    while timeout > 0:
        try:
            response = requests.get(ENDPOINT_URI, timeout=1)
            if response.status_code == 403:
                return
        except:  # pylint: disable=bare-except
            pass
        timeout -= 0.1
        time.sleep(0.1)
    raise RuntimeError("minio server didn't start")


@pytest.fixture()
def s3_base(tmpdir, pytestconfig) -> Iterator[None]:
    """Launch minio server."""
    if pytestconfig.getoption('s3') is False:
        pytest.skip('S3 disabled')
    if is_minio_up(timeout=1):
        raise RuntimeError('minio server already up')
    os.environ['MINIO_CACHE_AFTER'] = '1'
    os.environ['MINIO_CACHE'] = 'on'
    os.environ['MINIO_ROOT_PASSWORD'] = CREDENTIAL
    os.environ['MINIO_ROOT_USER'] = CREDENTIAL
    # pylint: disable=consider-using-with
    process = subprocess.Popen(
        shlex.split(f'minio server --quiet --address {ENDPOINT} '
                    f"--console-address :{PORT+1} '{tmpdir!s}'"))

    try:
        wait_for_minio_to_start(timeout=30)
        yield
    finally:
        process.terminate()
        process.wait()
    # pylint: enable=consider-using-with


def make_bucket(name) -> None:
    """Create a bucket."""
    session: botocore.session.Session = botocore.session.get_session()
    client = session.create_client(
        's3',
        aws_access_key_id=CREDENTIAL,
        aws_secret_access_key=CREDENTIAL,
        endpoint_url=ENDPOINT_URI,
        region_name='us-east-1',
        config=botocore.client.Config(signature_version='s3v4'))
    client.create_bucket(Bucket=name, ACL='public-read')


# pylint: disable=redefined-outer-name, unused-argument # pytest fixture
@pytest.fixture()
def s3(s3_base) -> Iterator[s3fs.core.S3FileSystem]:
    """Create a S3 file system instance."""
    s3fs.core.S3FileSystem.clear_instance_cache()
    fs = s3fs.core.S3FileSystem(anon=False,
                                key=CREDENTIAL,
                                secret=CREDENTIAL,
                                client_kwargs={'endpoint_url': ENDPOINT_URI})
    fs.invalidate_cache()
    yield fs
    # pylint: enable=redefined-outer-name, unused-argument


class S3Path(type(pathlib.Path())):  # type: ignore[misc]
    """Handle S3 path on multiple platforms."""

    def __str__(self) -> str:
        return super().__str__().replace('\\', '/')


class S3:
    """S3 filesystem."""
    #: Bucket ID
    ID = 0

    # pylint: disable=redefined-outer-name # pytest fixture
    def __init__(self, s3: s3fs.core.S3FileSystem) -> None:
        name: str = f'bucket{S3.ID}'
        S3.ID += 1
        make_bucket(name)
        self.collection: S3Path = S3Path(name).joinpath('collection')
        self.view: S3Path = S3Path(name).joinpath('view')
        self.fs: s3fs.core.S3FileSystem = s3

    # pylint: enable=redefined-outer-name



"""
Indexing a Collection.
======================

In this example, we will see how to index a collection.
"""
from typing import Iterator, List, Optional, Tuple, Union
import pathlib
import pprint

import dask.distributed
import fsspec
import numpy

import zcollection
import zcollection.indexing
import zcollection.partitioning.tests.data

# %%
# Initialization of the environment
# ---------------------------------
fs = fsspec.filesystem('memory')
cluster = dask.distributed.LocalCluster(processes=False)
client = dask.distributed.Client(cluster)

# %%
# A collection can be indexed. This allows quick access to the data without
# having to browse the entire dataset.
#
# Creating the test collection.
# -----------------------------
#
# For this latest example, we will index another data set. This one contains
# measurements of a fictitious satellite on several half-orbits.
zds: zcollection.Dataset = zcollection.Dataset.from_xarray(
    zcollection.partitioning.tests.data.create_test_sequence(5, 20, 10))
print(zds)

# %%
collection: zcollection.Collection = zcollection.create_collection(
    'time',
    zds,
    zcollection.partitioning.Date(('time', ), 'M'),
    partition_base_dir='/one_other_collection',
    filesystem=fs)
collection.insert(zds, merge_callable=zcollection.merging.merge_time_series)

# %%
# Here we have created a collection partitioned by month.
pprint.pprint(fs.listdir('/one_other_collection/year=2000'))


# %%
# Class to implement
# ------------------
#
# The idea of the implementation is to calculate for each visited partition, the
# slice of data that has a constant quantity. In our example, we will rely on
# the cycle and pass number information. The first method we will implement is
# the detection of these constant parts of two vectors containing the cycle and
# pass number.
def split_half_orbit(
    cycle_number: numpy.ndarray,
    pass_number: numpy.ndarray,
) -> Iterator[Tuple[int, int]]:
    """Calculate the indexes of the start and stop of each half-orbit.

    Args:
        pass_number: Pass numbers.
    Returns:
        Iterator of start and stop indexes.
    """
    assert pass_number.shape == cycle_number.shape
    pass_idx = numpy.where(numpy.roll(pass_number, 1) != pass_number)[0]
    cycle_idx = numpy.where(numpy.roll(cycle_number, 1) != cycle_number)[0]

    half_orbit = numpy.unique(
        numpy.concatenate(
            (pass_idx, cycle_idx, numpy.array([pass_number.size],
                                              dtype='int64'))))
    del pass_idx, cycle_idx

    yield from tuple(zip(half_orbit[:-1], half_orbit[1:]))


# %%
# Now we will compute these constant parts from a dataset contained in a
# partition.
def _half_orbit(
    zds: zcollection.Dataset,
    *args,
    **kwargs,
) -> numpy.ndarray:
    """Return the indexes of the start and stop of each half-orbit.

    Args:
        ds: Datasets stored in a partition to be indexed.
    Returns:
        Dictionary of start and stop indexes for each half-orbit.
    """
    pass_number_varname = kwargs.pop('pass_number', 'pass_number')
    cycle_number_varname = kwargs.pop('cycle_number', 'cycle_number')
    pass_number = zds.variables[pass_number_varname].values
    cycle_number = zds.variables[cycle_number_varname].values

    generator = ((
        i0,
        i1,
        cycle_number[i0],
        pass_number[i0],
    ) for i0, i1 in split_half_orbit(cycle_number, pass_number))

    return numpy.fromiter(generator, numpy.dtype(HalfOrbitIndexer.dtype()))


# %%
# Finally, we implement our indexing class. The base class
# (:py:class:`zcollection.indexing.Indexer<zcollection.indexing.abc.Indexer>`)
# implements the index update and the associated queries.
class HalfOrbitIndexer(zcollection.indexing.Indexer):
    """Index collection by half-orbit."""
    #: Column name of the cycle number.
    CYCLE_NUMBER = 'cycle_number'

    #: Column name of the pass number.
    PASS_NUMBER = 'pass_number'

    @classmethod
    def dtype(cls, /, **kwargs) -> List[Tuple[str, str]]:
        """Return the columns of the index.

        Returns:
            A tuple of (name, type) pairs.
        """
        return super().dtype() + [
            (cls.CYCLE_NUMBER, 'uint16'),
            (cls.PASS_NUMBER, 'uint16'),
        ]

    @classmethod
    def create(
        cls,
        path: Union[pathlib.Path, str],
        zds: zcollection.Collection,
        filesystem: Optional[fsspec.AbstractFileSystem] = None,
        **kwargs,
    ) -> 'HalfOrbitIndexer':
        """Create a new index.

        Args:
            path: The path to the index.
            ds: The collection to be indexed.
            filesystem: The filesystem to use.
        Returns:
            The created index.
        """
        return super()._create(path,
                               zds,
                               meta=dict(attribute=b'value'),
                               filesystem=filesystem)  # type: ignore

    def update(
        self,
        zds: zcollection.Collection,
        partition_size: Optional[int] = None,
        npartitions: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Update the index.

        Args:
            ds: New data stored in the collection to be indexed.
            partition_size: The length of each bag partition.
            npartitions: The number of desired bag partitions.
            cycle_number: The name of the cycle number variable stored in the
                collection. Defaults to "cycle_number".
            pass_number: The name of the pass number variable stored in the
                collection. Defaults to "pass_number".
        """
        super()._update(zds, _half_orbit, partition_size, npartitions,
                        **kwargs)


# %%
# Using the index
# ---------------
#
# Now we can create our index and fill it.
indexer: HalfOrbitIndexer = HalfOrbitIndexer.create('/index.parquet',
                                                    collection,
                                                    filesystem=fs)
indexer.update(collection)

# The following command allows us to view the information stored in our index:
# the first and last indexes of the partition associated with the registered
# half-orbit number and the identifier of the indexed partition.
indexer.table.to_pandas()

# %%
# This index can now be used to load a part of a collection.
selection: zcollection.Dataset | None = collection.load(
    indexer=indexer.query(dict(pass_number=[1, 2])),
    delayed=False,
)
assert selection is not None
selection.to_xarray()

# %%
# Close the local cluster to avoid printing warning messages in the other
# examples.
client.close()
cluster.close()



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Test the registry module.
=========================
"""
from typing import Any

import pytest

from .. import registry


def test_get_codecs() -> None:
    """Test the get_codecs function."""
    with pytest.raises(ValueError):
        registry.get_codecs({'ID': 'foo'})

    with pytest.raises(ValueError):
        registry.get_codecs({'id': 'foo'})


class MyCodec:
    """A dummy codec."""
    ID = 'foo'

    __slots__ = ('attribute', )

    def __init__(self, attribute) -> None:
        self.attribute: Any = attribute

    def get_config(self) -> dict:
        """Returns the configuration of the codec."""
        return {'id': self.ID, 'attribute': self.attribute}

    @classmethod
    def from_config(cls, config: dict) -> 'MyCodec':
        """Creates an instance from the given configuration."""
        return cls(config['attribute'])


def test_register_codec() -> None:
    """Test the register_codec function."""
    registry.register_codec(MyCodec, codec_id='foo')  # type: ignore[arg-type]

    instance = MyCodec(12)

    other = registry.get_codecs(instance.get_config())
    assert other.attribute == instance.attribute  # type: ignore
    assert isinstance(other, MyCodec)

    with pytest.raises(ValueError):
        registry.register_codec(
            MyCodec,  # type: ignore[arg-type]
            codec_id='foo')



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Fixtures for the tests.
=======================
"""
from typing import Literal

import pytest


@pytest.fixture
def dask_arrays() -> Literal[True]:
    """Load the data in Dask arrays."""
    return True


@pytest.fixture
def numpy_arrays() -> Literal[False]:
    """Load the data in NumPy arrays."""
    return False



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Convenience functions
=====================
"""
from __future__ import annotations

from typing import Literal

import xarray

from .. import collection, dataset, fs_utils, partitioning


def create_collection(
    axis: str,
    ds: xarray.Dataset | dataset.Dataset,
    partition_handler: partitioning.Partitioning,
    partition_base_dir: str,
    **kwargs,
) -> collection.Collection:
    """Create a collection.

    Args:
        axis: The axis to use for the collection.
        ds: The dataset to use.
        partition_handler: The partitioning handler to use.
        partition_base_dir: The base directory to use for the partitions.
        **kwargs: Additional parameters are passed through to the constructor
            of the class :py:class:`Collection`.

    Example:
        >>> import xarray as xr
        >>> import zcollection
        >>> data = xr.Dataset({
        ...     "a": xr.DataArray([1, 2, 3]),
        ...     "b": xr.DataArray([4, 5, 6])
        ... })
        >>> collection = zcollection.create_collection(
        ...     "a", data,
        ...     zcollection.partitioning.Sequence(("a", )),
        ...     "/tmp/my_collection")

    Returns:
        The collection.

    Raises:
        ValueError: If the base directory already exists.
    """
    filesystem = fs_utils.get_fs(kwargs.pop('filesystem', None))
    if filesystem.exists(partition_base_dir):
        raise ValueError(
            f'The directory {partition_base_dir!r} already exists.')
    if isinstance(ds, xarray.Dataset):
        ds = dataset.Dataset.from_xarray(ds)
    return collection.Collection(axis,
                                 ds.metadata(),
                                 partition_handler,
                                 partition_base_dir,
                                 mode='w',
                                 filesystem=filesystem,
                                 **kwargs)


# pylint: disable=redefined-builtin
def open_collection(path: str,
                    *,
                    mode: Literal['r', 'w'] | None = None,
                    **kwargs) -> collection.Collection:
    """Open a collection.

    Args:
        path: The path to the collection.
        mode: The mode to open the collection.
        **kwargs: Additional parameters are passed through the method
            :py:meth:`zcollection.collection.Collection.from_config`.
    Returns:
        The collection.

    Example:
        >>> import zcollection
        >>> collection = zcollection.open_collection(
        ...     "/tmp/mycollection", mode="r")
    """
    return collection.Collection.from_config(path, mode=mode, **kwargs)
    # pylint: enable=redefined-builtin



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""Test the time series merging."""
import copy

import numpy

from .. import time_series
from ...tests import data
# pylint: disable=unused-import # Need to import for fixtures
from ...tests.cluster import dask_client, dask_cluster

# pylint: enable=unused-import # Need to import for fixtures


def test_merge_disjoint(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test the update of two disjoint time series."""
    generator = data.create_test_dataset()
    zds0 = next(generator)
    zds1 = next(generator)

    zds = time_series.merge_time_series(zds1, zds0, 'time', 'num_lines')
    assert numpy.all(zds.variables['time'].values == numpy.concatenate((
        zds0.variables['time'].values, zds1.variables['time'].values)))

    zds = time_series.merge_time_series(zds0, zds1, 'time', 'num_lines')
    assert numpy.all(zds.variables['time'].values == numpy.concatenate((
        zds0.variables['time'].values, zds1.variables['time'].values)))

    zds = time_series.merge_time_series(zds0, zds0, 'time', 'num_lines')
    assert numpy.all(
        zds.variables['time'].values == zds0.variables['time'].values)


def test_merge_intersection(
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test the update of two intersecting time series."""
    generator = data.create_test_dataset()
    zds0 = next(generator)
    # ds0.variables["time"].values => numpy.array([
    #     "2000-01-01T00:00:00.000000", "2000-01-04T00:00:00.000000",
    #     "2000-01-07T00:00:00.000000", "2000-01-10T00:00:00.000000",
    #     "2000-01-13T00:00:00.000000", "2000-01-16T00:00:00.000000"
    # ])
    zds1 = next(generator)
    # ds1.variables["time"].values => numpy.array([
    #     "2000-01-19T00:00:00.000000", "2000-01-22T00:00:00.000000",
    #     "2000-01-25T00:00:00.000000", "2000-01-28T00:00:00.000000",
    #     "2000-01-31T00:00:00.000000"])

    existing_zds = zds1
    new_zds = copy.deepcopy(zds0)
    new_zds.variables['time'] = zds0.variables['time'].duplicate(
        zds0.variables['time'].values + numpy.timedelta64(9, 'D'))

    zds = time_series.merge_time_series(existing_zds, new_zds, 'time',
                                        'num_lines')
    assert numpy.all(zds.variables['time'].values == numpy.concatenate((
        zds0.variables['time'].values[3:], zds1.variables['time'].values[:])))

    existing_zds = zds0
    new_zds = copy.deepcopy(zds1)
    new_zds.variables['time'] = zds1.variables['time'].duplicate(
        zds1.variables['time'].values - numpy.timedelta64(9, 'D'))
    zds = time_series.merge_time_series(existing_zds, new_zds, 'time',
                                        'num_lines')
    assert numpy.all(zds.variables['time'].values == numpy.concatenate((
        zds0.variables['time'].values[:], zds1.variables['time'].values[:2])))

    existing_zds = zds0
    new_zds = zds0.isel({'num_lines': slice(1, -1)})
    new_zds.variables['var1'] = new_zds.variables['var1'].duplicate(
        new_zds.variables['var1'].values + 100)
    zds = time_series.merge_time_series(existing_zds, new_zds, 'time',
                                        'num_lines')
    assert numpy.all(zds.variables['var1'].values == numpy.concatenate((
        zds0.variables['var1'].values[:1],
        zds0.variables['var1'].values[1:-1] + 100,
        zds0.variables['var1'].values[-1:])))



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Testing the storage module.
===========================
"""
from typing import Any
import math
import platform
import time

import dask.array
import dask.distributed
import numpy
import pytest
import zarr

from .. import dataset, storage, sync
# pylint: disable=unused-import # Need to import for fixtures
from .cluster import dask_client, dask_cluster
from .fs import local_fs

# pylint: enable=unused-import


def test_execute_transaction(
        dask_client,  # pylint: disable=redefined-outer-name
) -> None:
    """Test the execute_transaction function."""
    # First case: no transaction to execute
    assert storage.execute_transaction(dask_client, sync.NoSync(), []) is None

    # General case: execute a transaction without error
    def func():
        return 1

    assert sum(
        storage.execute_transaction(
            dask_client, sync.NoSync(),
            [dask_client.submit(func) for i in range(10)])) == 10

    # Degraded case: execute a transaction with error
    def fail(data):
        time.sleep((10 - data) / 100)
        if data % 2 == 0:
            raise ValueError('Odd')
        return data

    exception_seen = False
    futures = []
    try:
        futures += storage.execute_transaction(
            dask_client, sync.NoSync(),
            [dask_client.submit(fail, i) for i in range(10)])
    except ValueError:
        exception_seen = True
    assert exception_seen
    for idx, item in enumerate(futures):
        if idx % 2 == 0:
            assert item.exception() is not None
        else:
            assert item.done()


def create_variable(shape, fill_value=None) -> dataset.DelayedArray:
    """Create a variable."""
    data = numpy.ones(shape, dtype='uint8')
    if fill_value is not None:
        data[:, 0] = fill_value
    return dataset.DelayedArray(name='var',
                                data=data,
                                dimensions=('x', 'y'),
                                fill_value=fill_value,
                                attrs=[
                                    dataset.Attribute('a', 1),
                                    dataset.Attribute('b', 2),
                                    dataset.Attribute('long_name', 'long name')
                                ])


def create_dataset(shape) -> dataset.Dataset:
    """Create a dataset."""
    return dataset.Dataset([create_variable(shape)],
                           attrs=[
                               dataset.Attribute('a', 1),
                               dataset.Attribute('b', 2),
                               dataset.Attribute('long_name', 'long name')
                           ])


def test_write_attrs(
        local_fs,  # pylint: disable=redefined-outer-name
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test the write_attrs function."""
    var = create_variable((10, 2))
    path = local_fs.root.joinpath('var')
    local_fs.mkdir(str(path))
    storage.write_zattrs(str(local_fs.root), var, local_fs.fs)
    path = str(path.joinpath(storage.ZATTRS))
    expected = [
        b'{\n',
        b'  "a": 1,\n',
        b'  "b": 2,\n',
        b'  "long_name": "long name",\n',
        b'  "_ARRAY_DIMENSIONS": [\n',
        b'    "x",\n',
        b'    "y"\n',
        b'  ]\n',
        b'}',
    ]
    if platform.platform().startswith('Windows'):
        expected = [item.replace(b'\n', b'\r\n') for item in expected]
    assert local_fs.exists(path)
    with local_fs.open(path) as stream:
        lines = stream.readlines()
        assert lines == expected
    assert local_fs.exists(str(local_fs.root.joinpath('var', storage.ZATTRS)))


@pytest.mark.parametrize('chunks, chunks_expected', [
    (None, (1024, 1024)),
    ({
        'x': -1,
        'y': -1
    }, (1024, 1024)),
    ({
        'x': -1,
        'y': 1024
    }, (1024, 1024)),
    ({
        'x': 1024,
        'y': -1
    }, (1024, 1024)),
    ({
        'x': 'auto',
        'y': 'auto'
    }, (1024, 1024)),
    ({
        'x': 512,
        'y': 'auto'
    }, (512, 1024)),
    ({
        'x': 256,
        'y': 512
    }, (256, 512)),
    ({
        'x': 53,
        'y': 826
    }, (53, 826)),
])
def test_write_variable(
        local_fs,  # pylint: disable=redefined-outer-name
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
        chunks,
        chunks_expected) -> None:
    """Test the write_variable function."""
    dim_size = 1024
    var = create_variable((dim_size, dim_size))
    storage.write_zarr_variable(('var', var),
                                dirname=str(local_fs.root),
                                fs=local_fs.fs,
                                chunks=chunks)
    path = str(local_fs.root.joinpath('var'))
    assert local_fs.exists(path)
    mapper = local_fs.get_mapper(path)
    zarray = zarr.open(mapper)
    assert zarray.shape == (dim_size, dim_size)
    assert numpy.all(zarray[...] == 1)

    chunks_number = math.ceil(dim_size / chunks_expected[0]) * math.ceil(
        dim_size / chunks_expected[1])
    assert zarray.chunks == chunks_expected
    assert zarray.nchunks == chunks_number

    other = storage.open_zarr_array(zarray, 'var')  # type:ignore
    assert other.metadata() == var.metadata()
    assert numpy.all(other.values == var.values)


@pytest.mark.parametrize(
    'block_size, chunks, blocks_number',
    [
        # Everything fit in one block
        (1048756, None, 1),
        # x and y are 'locked', only one block
        (1048756 // 20, {
            'x': -1,
            'y': -1
        }, 1),
        # Half block size, 2 blocks split along x
        (1048756 // 2, {
            'x': 512,
            'y': 'auto'
        }, 2),
        # Half block size, 2 blocks split along y
        (1048756 // 2, {
            'x': 1024,
            'y': 'auto'
        }, 2),
    ])
def test_write_variable_block_size_limit(
        local_fs,  # pylint: disable=redefined-outer-name
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
        block_size,
        chunks,
        blocks_number) -> None:
    """Test the write_variable function with different block size limits."""
    dim_size = 1024
    var = create_variable((dim_size, dim_size))
    storage.write_zarr_variable(('var', var),
                                dirname=str(local_fs.root),
                                fs=local_fs.fs,
                                chunks=chunks,
                                block_size_limit=block_size)
    path = str(local_fs.root.joinpath('var'))
    mapper = local_fs.get_mapper(path)
    zarray = zarr.open(mapper)

    assert zarray.nchunks == blocks_number


def test_write_zarr_group(
        local_fs,  # pylint: disable=redefined-outer-name
        dask_client,  # pylint: disable=redefined-outer-name
) -> None:
    """Test the write_zarr_group function."""
    zds = create_dataset((1024, 1024))
    # memory fs does not support multi-processes
    future = dask_client.submit(storage.write_zarr_group,
                                dask_client.scatter(zds), str(local_fs.root),
                                local_fs.fs, sync.NoSync())
    future.result()
    mapper = local_fs.get_mapper(str(local_fs.root))
    zarray = zarr.open_group(mapper)
    assert numpy.all(zarray['var'][...] == 1)
    assert zarray.attrs['a'] == 1
    assert zarray.attrs['b'] == 2
    assert zarray.attrs['long_name'] == 'long name'
    assert zarray['var'].attrs['_ARRAY_DIMENSIONS'] == ['x', 'y']

    other = storage.open_zarr_group(str(local_fs.root), local_fs.fs)
    assert other.metadata() == zds.metadata()


@pytest.mark.parametrize('chunks, chunks_expected', [
    (None, (1024, 1024)),
    ({
        'x': -1,
        'y': 1024
    }, (1024, 1024)),
    ({
        'x': 'auto',
        'y': 'auto'
    }, (1024, 1024)),
    ({
        'x': 512,
        'y': 'auto'
    }, (512, 1024)),
])
def test_update_zarr_array(
    local_fs,  # pylint: disable=redefined-outer-name
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    chunks,
    chunks_expected,
) -> None:
    """Test the update_zarr_array function."""
    dim_size = 1024

    var = create_variable((dim_size, dim_size), fill_value=10)
    storage.write_zarr_variable(('var', var),
                                str(local_fs.root),
                                local_fs.fs,
                                chunks=chunks)
    path = str(local_fs.root.joinpath('var'))
    storage.update_zarr_array(path, dask.array.full((dim_size, dim_size), 2),
                              local_fs.fs)
    mapper = local_fs.get_mapper(path)
    zarray = zarr.open(mapper)
    assert numpy.all(zarray[...] == 2)
    data: Any = numpy.full((dim_size, dim_size), 2)
    data[:, 0] = 5
    data = numpy.ma.masked_equal(data, 5)
    assert numpy.all(data[:, 0].mask)
    storage.update_zarr_array(path, data, local_fs.fs)
    zarray = zarr.open(mapper)
    assert numpy.all(zarray[:, 0] == 10)

    chunks_number = math.ceil(dim_size / chunks_expected[0]) * math.ceil(
        dim_size / chunks_expected[1])
    assert zarray.chunks == chunks_expected
    assert zarray.nchunks == chunks_number


def test_del_zarr_array(
        local_fs,  # pylint: disable=redefined-outer-name
        dask_client,  # pylint: disable=redefined-outer-name,unused-argument
) -> None:
    """Test the del_zarr_array function."""
    var = create_variable((1024, 1024))
    root = str(local_fs.root)
    storage.write_zarr_variable(('var', var), root, local_fs.fs)
    storage.del_zarr_array(root, 'var', local_fs.fs)
    assert not local_fs.exists(str(local_fs.root.joinpath('var')))


@pytest.mark.parametrize('chunks, chunks_expected', [
    (None, (1024, 1024)),
    ({
        'x': 1024,
        'y': 1024
    }, (1024, 1024)),
    ({
        'x': -1,
        'y': -1
    }, (1024, 1024)),
    ({
        'x': 512,
        'y': 256
    }, (512, 256)),
    ({
        'x': -1,
        'y': 128
    }, (1024, 128)),
])
def test_add_zarr_array(
    local_fs,  # pylint: disable=redefined-outer-name
    dask_client,  # pylint: disable=redefined-outer-name,unused-argument
    chunks,
    chunks_expected,
) -> None:
    """Test the add_zarr_array function."""
    dim_size = 1024
    var = create_variable((dim_size, dim_size), fill_value=10)
    root = str(local_fs.root)
    var.name = 'var1'
    storage.write_zarr_variable(('var1', var),
                                root,
                                local_fs.fs,
                                chunks=chunks)
    var.name = 'var2'
    storage.add_zarr_array(root,
                           var.metadata(),
                           'var1',
                           local_fs.fs,
                           chunks=chunks)
    mapper = local_fs.get_mapper(root)
    zarray = zarr.open(mapper)
    assert numpy.all(zarray['var2'][...] == 10)

    chunks_number = math.ceil(dim_size / chunks_expected[0]) * math.ceil(
        dim_size / chunks_expected[1])

    for varname in ['var1', 'var2']:
        var_array = zarray[varname]
        assert var_array.chunks == chunks_expected
        assert var_array.nchunks == chunks_number



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Variables of a dataset.
=======================
"""
from ..meta import Attribute
from .abc import Variable, new_variable
from .array import Array
from .delayed_array import DelayedArray

__all__ = ('Attribute', 'Variable', 'Array', 'DelayedArray', 'new_variable')



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Registers the partitioning codecs.
==================================
"""
from __future__ import annotations

from typing import Any

from . import abc

#: A registry of all available partitioning codecs.
CODEC_REGISTRY: dict[str, type[abc.Partitioning]] = {}


def get_codecs(config: dict[str, Any]) -> abc.Partitioning:
    """Get the partitioning scheme for the given configuration.

    Args:
        config: A dictionary of the partitioning configuration parameters.

    Returns:
        The partitioning scheme.

    Raises:
        ValueError: If the requested codec is not defined.
    """
    codec_id: Any | None = config.pop('id', None)
    if codec_id is None:
        raise ValueError(f'codec not available: {codec_id!r}')
    cls: type[abc.Partitioning] | None = CODEC_REGISTRY.get(codec_id, None)
    if cls is None:
        raise ValueError(f'codec not available: {codec_id!r}')
    return cls.from_config(config)


def register_codec(cls: type[abc.Partitioning],
                   *,
                   codec_id: str | None = None) -> None:
    """Register a partitioning scheme.

    Args:
        cls: The partitioning scheme class.
        codec_id: The partitioning scheme identifier.

    Raises:
        ValueError: If the codec identifier is already registered.
    """
    if codec_id is None:
        codec_id = cls.ID
    if codec_id is None:
        raise ValueError('codec identifier not defined')
    if codec_id in CODEC_REGISTRY:
        raise ValueError(f'codec already registered: {codec_id!r}')
    CODEC_REGISTRY[codec_id] = cls



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Convenience functions
=====================
"""
from .collection import create_collection, open_collection
from .view import create_view, open_view

__all__ = (
    'create_collection',
    'open_collection',
    'create_view',
    'open_view',
)



# Copyright (c) 2023 CNES
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Type hints for the zcollection package.
=======================================

.. rubric:: Type aliases

.. py:data:: DType
    :canonical: DType

    Type of a numpy array.

.. py:data:: DTypeLike
    :canonical: DTypeLike

    Type of a numpy array or a string.

.. py:data:: NDArray
    :canonical: NDArray

    Type of a numpy array.

.. py:data:: NDMaskedArray

    Type of a numpy array with a mask.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Protocol, TypeVar

try:
    from types import GenericAlias  # type: ignore[attr-defined]
except ImportError:
    # pylint: disable=ungrouped-imports
    # For Python < 3.9 we use a backport of GenericAlias provided by
    # numpy
    # isort: off
    from numpy._typing._generic_alias import (  # type: ignore[assignment,misc]
        _GenericAlias as GenericAlias,  # yapf: disable
    )
    # isort: on
    # pylint: enable=ungrouped-imports

try:
    from typing_extensions import TypeAlias
except ImportError:
    # pylint: disable=ungrouped-imports
    # TypeAlias is defined in typing starting from 3.10
    from typing import TypeAlias  # type: ignore[attr-defined,no-redef]
    # pylint: enable=ungrouped-imports

import numpy
import numpy.typing

# pylint: disable=invalid-name
_DType_co = TypeVar('_DType_co', covariant=True, bound='numpy.dtype[Any]')
_ScalarType_co = TypeVar('_ScalarType_co', bound=numpy.generic, covariant=True)
# pylint: enable=invalid-name

if TYPE_CHECKING:
    DType = numpy.dtype[_ScalarType_co]
    NDMaskedArray = numpy.ma.MaskedArray[Any, DType]  # pragma: no cover
else:
    DType = GenericAlias(numpy.dtype, (_ScalarType_co, ))
    NDMaskedArray = GenericAlias(numpy.ma.MaskedArray, (Any, DType))

NDArray: TypeAlias = numpy.typing.NDArray  # pragma: no cover
DTypeLike: TypeAlias = numpy.typing.DTypeLike  # pragma: no cover


class ArrayLike(Protocol[_DType_co]):
    """Protocol for array-like objects."""

    def __array__(self) -> NDArray:
        ...

    @property
    def dtype(self) -> DType:
        """The data type of the array."""
        # pylint: disable=unnecessary-ellipsis
        # Make checker happy.
        ...
        # pylint: enable=unnecessary-ellipsis

    @property
    def shape(self) -> tuple[int, ...]:
        """The shape of the array."""
        # pylint: disable=unnecessary-ellipsis
        # Make checker happy.
        ...
        # pylint: enable=unnecessary-ellipsis

    @property
    def size(self) -> int:
        """The size of the array."""
        # pylint: disable=unnecessary-ellipsis
        # Make checker happy.
        ...
        # pylint: enable=unnecessary-ellipsis

    def astype(self, dtype: DTypeLike) -> ArrayLike[_DType_co]:
        """Convert the array to a given type."""
        # pylint: disable=unnecessary-ellipsis
        # Make checker happy.
        ...
        # pylint: enable=unnecessary-ellipsis
